import copy
import datetime
import logging
import math
import re
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from itertools import count, repeat
from typing import Optional
from urllib.parse import urljoin, urlparse, urlsplit, urlunparse, urlunsplit

from isodate import Duration, UTC as utc, parse_datetime, parse_duration  # type: ignore[import]


log = logging.getLogger(__name__)
epoch_start = datetime.datetime(1970, 1, 1, tzinfo=utc)


class Segment:
    def __init__(self, url, duration, init=False, content=True, available_at=epoch_start, range=None):
        self.url = url
        self.duration = duration
        self.init = init
        self.content = content
        self.available_at = available_at
        self.range = range


def datetime_to_seconds(dt):
    return (dt - epoch_start).total_seconds()


def count_dt(firstval=datetime.datetime.now(tz=utc), step=datetime.timedelta(seconds=1)):
    x = firstval
    while True:
        yield x
        x += step


@contextmanager
def freeze_timeline(mpd):
    timelines = copy.copy(mpd.timelines)
    yield
    mpd.timelines = timelines


class MPDParsers:
    @staticmethod
    def bool_str(v):
        return v.lower() == "true"

    @staticmethod
    def type(type_):
        if type_ not in ("static", "dynamic"):
            raise MPDParsingError("@type must be static or dynamic")
        return type_

    @staticmethod
    def duration(duration):
        return parse_duration(duration)

    @staticmethod
    def datetime(dt):
        return parse_datetime(dt).replace(tzinfo=utc)

    @staticmethod
    def segment_template(url_template):
        end = 0
        res = ""
        for m in re.compile(r"(.*?)\$(\w+)(?:%([\w.]+))?\$").finditer(url_template):
            _, end = m.span()
            res += "{0}{{{1}{2}}}".format(m.group(1), m.group(2), f":{m.group(3)}" if m.group(3) else "")

        return (res + url_template[end:]).format

    @staticmethod
    def frame_rate(frame_rate):
        if "/" not in frame_rate:
            return float(frame_rate)

        a, b = frame_rate.split("/")
        return float(a) / float(b)

    @staticmethod
    def timedelta(timescale=1):
        def _timedelta(seconds):
            return datetime.timedelta(seconds=int(float(seconds) / float(timescale)))

        return _timedelta

    @staticmethod
    def range(range_spec):
        r = range_spec.split("-")
        if len(r) != 2:
            raise MPDParsingError("invalid byte-range-spec")

        start, end = int(r[0]), r[1] and int(r[1]) or None
        return start, end and ((end - start) + 1)


class MPDParsingError(Exception):
    pass


class MPDNode:
    __tag__: Optional[str] = None

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        self.node = node
        self.root = root
        self.parent = parent
        self._base_url = kwargs.get("base_url")
        self.attributes = set()
        if self.__tag__ and self.node.tag.lower() != self.__tag__.lower():
            raise MPDParsingError("root tag did not match the expected tag: {}".format(self.__tag__))

    @property
    def attrib(self):
        return self.node.attrib

    @property
    def text(self):
        return self.node.text

    def __str__(self):
        return "<{tag} {attrs}>".format(
            tag=self.__tag__,
            attrs=" ".join("@{}={}".format(attr, getattr(self, attr)) for attr in self.attributes)
        )

    def attr(self, key, default=None, parser=None, required=False, inherited=False):
        self.attributes.add(key)
        if key in self.attrib:
            value = self.attrib.get(key)
            if parser and callable(parser):
                return parser(value)
            else:
                return value
        elif inherited:
            if self.parent and hasattr(self.parent, key) and getattr(self.parent, key):
                return getattr(self.parent, key)

        if required:
            raise MPDParsingError("could not find required attribute {tag}@{attr} ".format(attr=key, tag=self.__tag__))
        else:
            return default

    def children(self, cls, minimum=0, maximum=None):

        children = self.node.findall(cls.__tag__)
        if len(children) < minimum or (maximum and len(children) > maximum):
            raise MPDParsingError("expected to find {}/{} required [{}..{})".format(
                self.__tag__, cls.__tag__, minimum, maximum or "unbound"))

        return list(map(lambda x: cls(x[1], root=self.root, parent=self, i=x[0], base_url=self.base_url),
                        enumerate(children)))

    def only_child(self, cls, minimum=0):
        children = self.children(cls, minimum=minimum, maximum=1)
        return children[0] if len(children) else None

    def walk_back(self, cls=None, f=lambda x: x):
        node = self.parent
        while node:
            if cls is None or cls.__tag__ == node.__tag__:
                yield f(node)
            node = node.parent

    def walk_back_get_attr(self, attr):
        parent_attrs = [getattr(n, attr) for n in self.walk_back() if hasattr(n, attr)]
        return parent_attrs[0] if len(parent_attrs) else None

    @property
    def base_url(self):
        base_url = self._base_url
        if hasattr(self, "baseURLs") and len(self.baseURLs):
            base_url = BaseURL.join(base_url, self.baseURLs[0].url)
        return base_url


class MPD(MPDNode):
    """
    Represents the MPD as a whole

    Should validate the XML input and provide methods to get segment URLs for each Period, AdaptationSet and
    Representation.

    """
    __tag__ = "MPD"

    def __init__(self, node, root=None, parent=None, url=None, *args, **kwargs):
        # top level has no parent
        super().__init__(node, root=self, *args, **kwargs)
        # parser attributes
        self.url = url
        self.timelines = defaultdict(lambda: -1)
        self.timelines.update(kwargs.pop("timelines", {}))
        self.id = self.attr("id")
        self.profiles = self.attr("profiles", required=True)
        self.type = self.attr("type", default="static", parser=MPDParsers.type)
        self.minimumUpdatePeriod = self.attr("minimumUpdatePeriod", parser=MPDParsers.duration, default=Duration())
        self.minBufferTime = self.attr("minBufferTime", parser=MPDParsers.duration, required=True)
        self.timeShiftBufferDepth = self.attr("timeShiftBufferDepth", parser=MPDParsers.duration)
        self.availabilityStartTime = self.attr("availabilityStartTime", parser=MPDParsers.datetime,
                                               default=datetime.datetime.fromtimestamp(0, utc),  # earliest date
                                               required=self.type == "dynamic")
        self.publishTime = self.attr("publishTime", parser=MPDParsers.datetime, required=self.type == "dynamic")
        self.availabilityEndTime = self.attr("availabilityEndTime", parser=MPDParsers.datetime)
        self.mediaPresentationDuration = self.attr("mediaPresentationDuration", parser=MPDParsers.duration)
        self.suggestedPresentationDelay = self.attr("suggestedPresentationDelay", parser=MPDParsers.duration)

        # parse children
        location = self.children(Location)
        self.location = location[0] if location else None
        if self.location:
            self.url = self.location.text
            urlp = list(urlparse(self.url))
            if urlp[2]:
                urlp[2], _ = urlp[2].rsplit("/", 1)
            self._base_url = urlunparse(urlp)

        self.baseURLs = self.children(BaseURL)
        self.periods = self.children(Period, minimum=1)
        self.programInformation = self.children(ProgramInformation)


class ProgramInformation(MPDNode):
    __tag__ = "ProgramInformation"


class BaseURL(MPDNode):
    __tag__ = "BaseURL"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)
        self.url = self.node.text.strip()

    @property
    def is_absolute(self):
        return urlparse(self.url).scheme

    @staticmethod
    def join(url, other):
        # if the other URL is an absolute url, then return that
        if urlparse(other).scheme:
            return other
        elif url:
            parts = list(urlsplit(url))
            if not parts[2].endswith("/"):
                parts[2] += "/"
            url = urlunsplit(parts)
            return urljoin(url, other)
        else:
            return other


class Location(MPDNode):
    __tag__ = "Location"


class Period(MPDNode):
    __tag__ = "Period"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)
        self.i = kwargs.get("i", 0)
        self.id = self.attr("id")
        self.bitstreamSwitching = self.attr("bitstreamSwitching", parser=MPDParsers.bool_str)
        self.duration = self.attr("duration", default=Duration(), parser=MPDParsers.duration)
        self.start = self.attr("start", default=Duration(), parser=MPDParsers.duration)

        if self.start is None and self.i == 0 and self.root.type == "static":
            self.start = 0

        # TODO: Early Access Periods

        self.baseURLs = self.children(BaseURL)
        self.segmentBase = self.only_child(SegmentBase)
        self.adaptationSets = self.children(AdaptationSet, minimum=1)
        self.segmentList = self.only_child(SegmentList)
        self.segmentTemplate = self.only_child(SegmentTemplate)
        self.sssetIdentifier = self.only_child(AssetIdentifier)
        self.eventStream = self.children(EventStream)
        self.subset = self.children(Subset)


class SegmentBase(MPDNode):
    __tag__ = "SegmentBase"


class AssetIdentifier(MPDNode):
    __tag__ = "AssetIdentifier"


class Subset(MPDNode):
    __tag__ = "Subset"


class EventStream(MPDNode):
    __tag__ = "EventStream"


class Initialization(MPDNode):
    __tag__ = "Initialization"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)
        self.source_url = self.attr("sourceURL")


class SegmentURL(MPDNode):
    __tag__ = "SegmentURL"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)
        self.media = self.attr("media")
        self.media_range = self.attr("mediaRange", parser=MPDParsers.range)


class SegmentList(MPDNode):
    __tag__ = "SegmentList"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)

        self.presentation_time_offset = self.attr("presentationTimeOffset")
        self.timescale = self.attr("timescale", parser=int)
        self.duration = self.attr("duration", parser=int)
        self.start_number = self.attr("startNumber", parser=int, default=1)

        if self.duration:
            self.duration_seconds = self.duration / float(self.timescale)
        else:
            self.duration_seconds = None

        self.initialization = self.only_child(Initialization)
        self.segment_urls = self.children(SegmentURL, minimum=1)

    @property
    def segments(self):
        if self.initialization:
            yield Segment(self.make_url(self.initialization.source_url), 0, init=True, content=False)
        for n, segment_url in enumerate(self.segment_urls, self.start_number):
            yield Segment(self.make_url(segment_url.media), self.duration_seconds, range=segment_url.media_range)

    def make_url(self, url):
        return BaseURL.join(self.base_url, url)


class AdaptationSet(MPDNode):
    __tag__ = "AdaptationSet"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)

        self.id = self.attr("id")
        self.group = self.attr("group")
        self.mimeType = self.attr("mimeType")
        self.lang = self.attr("lang")
        self.contentType = self.attr("contentType")
        self.par = self.attr("par")
        self.minBandwidth = self.attr("minBandwidth")
        self.maxBandwidth = self.attr("maxBandwidth")
        self.minWidth = self.attr("minWidth", parser=int)
        self.maxWidth = self.attr("maxWidth", parser=int)
        self.minHeight = self.attr("minHeight", parser=int)
        self.maxHeight = self.attr("maxHeight", parser=int)
        self.minFrameRate = self.attr("minFrameRate", parser=MPDParsers.frame_rate)
        self.maxFrameRate = self.attr("maxFrameRate", parser=MPDParsers.frame_rate)
        self.segmentAlignment = self.attr("segmentAlignment", default=False, parser=MPDParsers.bool_str)
        self.bitstreamSwitching = self.attr("bitstreamSwitching", parser=MPDParsers.bool_str)
        self.subsegmentAlignment = self.attr("subsegmentAlignment", default=False, parser=MPDParsers.bool_str)
        self.subsegmentStartsWithSAP = self.attr("subsegmentStartsWithSAP", default=0, parser=int)

        self.baseURLs = self.children(BaseURL)
        self.segmentTemplate = self.only_child(SegmentTemplate)
        self.representations = self.children(Representation, minimum=1)
        self.contentProtection = self.children(ContentProtection)


class SegmentTemplate(MPDNode):
    __tag__ = "SegmentTemplate"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)
        self.defaultSegmentTemplate = self.walk_back_get_attr('segmentTemplate')

        self.initialization = self.attr("initialization", parser=MPDParsers.segment_template)
        self.media = self.attr("media", parser=MPDParsers.segment_template)
        self.duration = self.attr("duration", parser=int,
                                  default=self.defaultSegmentTemplate.duration if self.defaultSegmentTemplate else None)
        self.timescale = self.attr("timescale", parser=int,
                                   default=self.defaultSegmentTemplate.timescale if self.defaultSegmentTemplate else 1)
        self.startNumber = self.attr("startNumber", parser=int,
                                     default=self.defaultSegmentTemplate.startNumber if self.defaultSegmentTemplate else 1)
        self.presentationTimeOffset = self.attr("presentationTimeOffset", parser=MPDParsers.timedelta(self.timescale))

        if self.duration:
            self.duration_seconds = self.duration / float(self.timescale)
        else:
            self.duration_seconds = None

        self.period = list(self.walk_back(Period))[0]

        # children
        self.segmentTimeline = self.only_child(SegmentTimeline)

    def segments(self, **kwargs):
        if kwargs.pop("init", True):
            init_url = self.format_initialization(**kwargs)
            if init_url:
                yield Segment(init_url, 0, True, False)
        for media_url, available_at in self.format_media(**kwargs):
            yield Segment(media_url, self.duration_seconds, False, True, available_at)

    def make_url(self, url):
        """
        Join the URL with the base URL, unless it's an absolute URL
        :param url: maybe relative URL
        :return: joined URL
        """
        return BaseURL.join(self.base_url, url)

    def format_initialization(self, **kwargs):
        if self.initialization:
            return self.make_url(self.initialization(**kwargs))

    def segment_numbers(self):
        """
        yield the segment number and when it will be available
        There are two cases for segment number generation, static and dynamic.

        In the case of static stream, the segment number starts at the startNumber and counts
        up to the number of segments that are represented by the periods duration.

        In the case of dynamic streams, the segments should appear at the specified time
        in the simplest case the segment number is based on the time since the availabilityStartTime
        :return:
        """
        log.debug("Generating segment numbers for {0} playlist (id={1})".format(self.root.type, self.parent.id))
        if self.root.type == "static":
            available_iter = repeat(epoch_start)
            duration = self.period.duration.seconds or self.root.mediaPresentationDuration.seconds
            if duration:
                number_iter = range(self.startNumber, int(duration / self.duration_seconds) + 1)
            else:
                number_iter = count(self.startNumber)
        else:
            now = datetime.datetime.now(utc)
            if self.presentationTimeOffset:
                since_start = (now - self.presentationTimeOffset) - self.root.availabilityStartTime
                available_start_date = self.root.availabilityStartTime + self.presentationTimeOffset + since_start
                available_start = available_start_date
            else:
                since_start = now - self.root.availabilityStartTime
                available_start = now

            # if there is no delay, use a delay of 3 seconds
            suggested_delay = datetime.timedelta(seconds=(self.root.suggestedPresentationDelay.total_seconds()
                                                          if self.root.suggestedPresentationDelay
                                                          else 3))

            # the number of the segment that is available at NOW - SUGGESTED_DELAY - BUFFER_TIME
            number_iter = count(
                self.startNumber
                + int(
                    (since_start - suggested_delay - self.root.minBufferTime).total_seconds()
                    / self.duration_seconds
                )
            )

            # the time the segment number is available at NOW
            available_iter = count_dt(available_start,
                                      step=datetime.timedelta(seconds=self.duration_seconds))

        yield from zip(number_iter, available_iter)

    def format_media(self, **kwargs):
        if self.segmentTimeline:
            if self.parent.id is None:
                # workaround for invalid `self.root.timelines[self.parent.id]`
                # creates a timeline for every mimeType instead of one for both
                self.parent.id = self.parent.mimeType
            log.debug("Generating segment timeline for {0} playlist (id={1}))".format(self.root.type, self.parent.id))
            if self.root.type == "dynamic":
                # if there is no delay, use a delay of 3 seconds
                suggested_delay = datetime.timedelta(seconds=(self.root.suggestedPresentationDelay.total_seconds()
                                                              if self.root.suggestedPresentationDelay
                                                              else 3))
                publish_time = self.root.publishTime or epoch_start

                # transform the time line in to a segment list
                timeline = []
                available_at = publish_time
                for segment, n in reversed(list(zip(self.segmentTimeline.segments, count(self.startNumber)))):
                    # the last segment in the timeline is the most recent
                    # so, work backwards and calculate when each of the segments was
                    # available, based on the durations relative to the publish time
                    url = self.make_url(self.media(Time=segment.t, Number=n, **kwargs))
                    duration = datetime.timedelta(seconds=segment.d / self.timescale)

                    # once the suggested_delay is reach stop
                    if self.root.timelines[self.parent.id] == -1 and publish_time - available_at >= suggested_delay:
                        break

                    timeline.append((url, available_at, segment.t))

                    available_at -= duration  # walk backwards in time

                # return the segments in chronological order
                for url, available_at, t in reversed(timeline):
                    if t > self.root.timelines[self.parent.id]:
                        self.root.timelines[self.parent.id] = t
                        yield (url, available_at)

            else:
                for segment, n in zip(self.segmentTimeline.segments, count(self.startNumber)):
                    yield (self.make_url(self.media(Time=segment.t, Number=n, **kwargs)),
                           datetime.datetime.now(tz=utc))

        else:
            for number, available_at in self.segment_numbers():
                yield (self.make_url(self.media(Number=number, **kwargs)),
                       available_at)


class Representation(MPDNode):
    __tag__ = "Representation"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)
        self.id = self.attr("id", required=True)
        self.bandwidth = self.attr("bandwidth", parser=lambda b: float(b) / 1000.0, required=True)
        self.mimeType = self.attr("mimeType", required=True, inherited=True)

        self.codecs = self.attr("codecs")
        self.startWithSAP = self.attr("startWithSAP")

        # video
        self.width = self.attr("width", parser=int)
        self.height = self.attr("height", parser=int)
        self.frameRate = self.attr("frameRate", parser=MPDParsers.frame_rate)

        # audio
        self.audioSamplingRate = self.attr("audioSamplingRate", parser=int)
        self.numChannels = self.attr("numChannels", parser=int)

        # subtitle
        self.lang = self.attr("lang", inherited=True)

        self.baseURLs = self.children(BaseURL)
        self.subRepresentation = self.children(SubRepresentation)
        self.segmentBase = self.only_child(SegmentBase)
        self.segmentList = self.children(SegmentList)
        self.segmentTemplate = self.only_child(SegmentTemplate)

    @property
    def bandwidth_rounded(self):
        return round(self.bandwidth, 1 - int(math.log10(self.bandwidth)))

    def segments(self, **kwargs):
        """
        Segments are yielded when they are available

        Segments appear on a time line, for dynamic content they are only available at a certain time
        and sometimes for a limited time. For static content they are all available at the same time.

        :param kwargs: extra args to pass to the segment template
        :return: yields Segments
        """

        # segmentBase = self.segmentBase or self.walk_back_get_attr("segmentBase")
        segmentLists = self.segmentList or self.walk_back_get_attr("segmentList")
        segmentTemplate = self.segmentTemplate or self.walk_back_get_attr("segmentTemplate")

        if segmentTemplate:
            for segment in segmentTemplate.segments(RepresentationID=self.id,
                                                    Bandwidth=int(self.bandwidth * 1000),
                                                    **kwargs):
                if segment.init:
                    yield segment
                else:
                    yield segment
        elif segmentLists:
            for segmentList in segmentLists:
                yield from segmentList.segments
        else:
            yield Segment(self.base_url, 0, True, True)


class SubRepresentation(MPDNode):
    __tag__ = "SubRepresentation"


class SegmentTimeline(MPDNode):
    __tag__ = "SegmentTimeline"
    TimelineSegment = namedtuple("TimelineSegment", "t d")

    def __init__(self, node, *args, **kwargs):
        super().__init__(node, *args, **kwargs)

        self.timescale = self.walk_back_get_attr("timescale")

        self.timeline_segments = self.children(_TimelineSegment)

    @property
    def segments(self):
        t = 0
        for tsegment in self.timeline_segments:
            if t == 0 and tsegment.t is not None:
                t = tsegment.t
            # check the start time from MPD
            for repeated_i in range(tsegment.r + 1):
                yield self.TimelineSegment(t, tsegment.d)
                t += tsegment.d


class _TimelineSegment(MPDNode):
    __tag__ = "S"

    def __init__(self, node, *args, **kwargs):
        super().__init__(node, *args, **kwargs)

        self.t = self.attr("t", parser=int)
        self.d = self.attr("d", parser=int)
        self.r = self.attr("r", parser=int, default=0)


class ContentProtection(MPDNode):
    __tag__ = "ContentProtection"

    def __init__(self, node, root=None, parent=None, *args, **kwargs):
        super().__init__(node, root, parent, *args, **kwargs)

        self.schemeIdUri = self.attr("schemeIdUri")
        self.value = self.attr("value")
        self.default_KID = self.attr("default_KID")



"""
$description Video streaming service focused on anime, manga, and dorama.
$url crunchyroll.com
$type vod
"""

import datetime
import logging
import re
from uuid import uuid4

from streamlink.plugin import Plugin, PluginError, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)

STREAM_WEIGHTS = {
    "low": 240,
    "mid": 420,
    "high": 720,
    "ultra": 1080,
}
STREAM_NAMES = {
    "120k": "low",
    "328k": "mid",
    "864k": "high"
}


def parse_timestamp(ts):
    """Takes ISO 8601 format(string) and converts into a utc datetime(naive)"""
    return (
        datetime.datetime.strptime(ts[:-7], "%Y-%m-%dT%H:%M:%S")
        + datetime.timedelta(hours=int(ts[-5:-3]), minutes=int(ts[-2:]))
        * int(f"{ts[-6:-5]}1")
    )


_api_schema = validate.Schema({
    "error": bool,
    validate.optional("code"): str,
    validate.optional("message"): str,
    validate.optional("data"): object,
})
_media_schema = validate.Schema(
    {
        validate.optional("name"): validate.any(str, None),
        validate.optional("series_name"): validate.any(str, None),
        validate.optional("media_type"): validate.any(str, None),
        "stream_data": validate.any(
            None,
            {
                "streams": validate.all(
                    [{
                        "quality": validate.any(str, None),
                        "url": validate.url(
                            scheme="http",
                            path=validate.endswith(".m3u8")
                        ),
                        validate.optional("video_encode_id"): str
                    }]
                )
            }
        )
    }
)
_login_schema = validate.Schema({
    "auth": validate.any(str, None),
    "expires": validate.all(
        str,
        validate.transform(parse_timestamp)
    ),
    "user": {
        "username": validate.any(str, None),
        "email": str,
    }
})
_session_schema = validate.Schema(
    {
        "session_id": str,
    },
    validate.get("session_id")
)


class CrunchyrollAPIError(Exception):
    """Exception thrown by the Crunchyroll API when an error occurs"""

    def __init__(self, msg, code):
        Exception.__init__(self, msg)
        self.msg = msg
        self.code = code


class CrunchyrollAPI:
    _api_url = "https://api.crunchyroll.com/{0}.0.json"
    _default_locale = "en_US"
    _version_name = "1.3.1.0"
    _access_token = "LNDJgOit5yaRIWN"
    _access_type = "com.crunchyroll.windows.desktop"

    def __init__(self, cache, session, session_id=None, locale=_default_locale):
        """Abstract the API to access to Crunchyroll data.

        Can take saved credentials to use on its calls to the API.
        """
        self.cache = cache
        self.session = session
        self.session_id = session_id
        if self.session_id:  # if the session ID is setup don't use the cached auth token
            self.auth = None
        else:
            self.auth = cache.get("auth")
        self.device_id = cache.get("device_id") or self.generate_device_id()
        self.locale = locale

    def _api_call(self, entrypoint, params=None, schema=None):
        """Makes a call against the api.

        :param entrypoint: API method to call.
        :param params: parameters to include in the request data.
        :param schema: schema to use to validate the data
        """
        url = self._api_url.format(entrypoint)

        # Default params
        params = params or {}
        if self.session_id:
            params.update({
                "session_id": self.session_id
            })
        else:
            params.update({
                "device_id": self.device_id,
                "device_type": self._access_type,
                "access_token": self._access_token,
            })
        params.update({
            "locale": self.locale.replace("_", ""),
            "version": self._version_name,
            "connectivity_type": "ethernet",
        })

        if self.session_id:
            params["session_id"] = self.session_id

        res = self.session.http.post(url, data=params)
        json_res = self.session.http.json(res, schema=_api_schema)

        if json_res["error"]:
            err_msg = json_res.get("message", "Unknown error")
            err_code = json_res.get("code", "unknown_error")
            raise CrunchyrollAPIError(err_msg, err_code)

        data = json_res.get("data")
        if schema:
            data = schema.validate(data, name="API response")

        return data

    def generate_device_id(self):
        device_id = str(uuid4())
        # cache the device id
        self.cache.set("device_id", device_id, expires=365 * 24 * 60 * 60)
        log.debug("Device ID: {0}".format(device_id))
        return device_id

    def start_session(self):
        """
            Starts a session against Crunchyroll's server.
            Is recommended that you call this method before making any other calls
            to make sure you have a valid session against the server.
        """
        params = {}
        if self.auth:
            params["auth"] = self.auth
        self.session_id = self._api_call("start_session", params, schema=_session_schema)
        log.debug("Session created with ID: {0}".format(self.session_id))
        return self.session_id

    def login(self, username, password):
        """
            Authenticates the session to be able to access restricted data from
            the server (e.g. premium restricted videos).
        """
        params = {
            "account": username,
            "password": password
        }

        login = self._api_call("login", params, schema=_login_schema)
        self.auth = login["auth"]
        self.cache.set("auth", login["auth"], expires_at=login["expires"])
        return login

    def authenticate(self):
        try:
            data = self._api_call("authenticate", {"auth": self.auth}, schema=_login_schema)
        except CrunchyrollAPIError:
            self.auth = None
            self.cache.set("auth", None, expires=0)
            log.warning("Saved credentials have expired")
            return

        log.debug("Credentials expire at: {}".format(data["expires"]))
        self.cache.set("auth", self.auth, expires_at=data["expires"])
        return data

    def get_info(self, media_id, fields=None, schema=None):
        """
            Returns the data for a certain media item.

            :param media_id: id that identifies the media item to be accessed.
            :param fields: list of the media"s field to be returned. By default the
            API returns some fields, but others are not returned unless they are
            explicity asked for. I have no real documentation on the fields, but
            they all seem to start with the "media." prefix (e.g. media.name,
            media.stream_data).
            :param schema: validation schema to use
        """
        params = {"media_id": media_id}

        if fields:
            params["fields"] = ",".join(fields)

        return self._api_call("info", params, schema=schema)


@pluginmatcher(re.compile(r"""
    https?://(\w+\.)?crunchyroll\.
    (?:
        com|de|es|fr|co\.jp
    )
    (?:
        /(en-gb|es|es-es|pt-pt|pt-br|fr|de|ar|it|ru)
    )?
    (?:
        (?:
            (?:/[^/&?]+)?
            /[^/&?]+-(?P<media_id>\d+)
        )
        |
        /watch/(?P<beta_id>\w+)/[\w-]+
    )
""", re.VERBOSE))
@pluginargument(
    "username",
    requires=["password"],
    metavar="USERNAME",
    help="A Crunchyroll username to allow access to restricted streams.",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    nargs="?",
    const=None,
    default=None,
    help="""
        A Crunchyroll password for use with --crunchyroll-username.

        If left blank you will be prompted.
    """,
)
@pluginargument(
    "purge-credentials",
    action="store_true",
    help="Purge cached Crunchyroll credentials to initiate a new session and reauthenticate.",
)
@pluginargument(
    "session-id",
    sensitive=True,
    metavar="SESSION_ID",
    help="""
        Set a specific session ID for crunchyroll, can be used to bypass
        region restrictions. If using an authenticated session ID, it is
        recommended that the authentication parameters be omitted as the
        session ID is account specific.

        Note: The session ID will be overwritten if authentication is used
        and the session ID does not match the account.
    """,
)
class Crunchyroll(Plugin):
    @classmethod
    def stream_weight(cls, key):
        weight = STREAM_WEIGHTS.get(key)
        if weight:
            return weight, "crunchyroll"

        return Plugin.stream_weight(key)

    def _get_streams(self):
        beta_id = self.match.group("beta_id")
        if beta_id:
            json = self.session.http.get(self.url, schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(), 'window.__INITIAL_STATE__')]/text()"),
                validate.none_or_all(
                    re.compile(r"window.__INITIAL_STATE__\s*=\s*({.*});"),
                    validate.none_or_all(
                        validate.get(1),
                        validate.parse_json(),
                        validate.none_or_all(
                            {"content": {"byId": {str: {"external_id": validate.all(
                                validate.transform(lambda s: int(s.replace("EPI.", ""))),
                            )}}}},
                            validate.get(("content", "byId")),
                        ),
                    ),
                ),
            ))
            if not json or beta_id not in json:
                return
            media_id = json[beta_id]["external_id"]
        else:
            media_id = int(self.match.group("media_id"))

        api = self._create_api()
        try:
            # the media.stream_data field is required, no stream data is returned otherwise
            info = api.get_info(media_id, fields=["media.name", "media.series_name",
                                "media.media_type", "media.stream_data"], schema=_media_schema)
        except CrunchyrollAPIError as err:
            raise PluginError(f"Media lookup error: {err.msg}")

        if not info:
            return

        streams = {}

        self.id = media_id
        self.title = info.get("name")
        self.author = info.get("series_name")
        self.category = info.get("media_type")

        info = info["stream_data"]

        # The adaptive quality stream sometimes a subset of all the other streams listed, ultra is no included
        has_adaptive = any(s["quality"] == "adaptive" for s in info["streams"])
        if has_adaptive:
            log.debug("Loading streams from adaptive playlist")
            for stream in filter(lambda x: x["quality"] == "adaptive", info["streams"]):
                for q, s in HLSStream.parse_variant_playlist(self.session, stream["url"]).items():
                    # rename the bitrates to low, mid, or high. ultra doesn't seem to appear in the adaptive streams
                    name = STREAM_NAMES.get(q, q)
                    streams[name] = s

        # If there is no adaptive quality stream then parse each individual result
        for stream in info["streams"]:
            if stream["quality"] != "adaptive":
                # the video_encode_id indicates that the stream is not a variant playlist
                if "video_encode_id" in stream:
                    streams[stream["quality"]] = HLSStream(self.session, stream["url"])
                else:
                    # otherwise the stream url is actually a list of stream qualities
                    for q, s in HLSStream.parse_variant_playlist(self.session, stream["url"]).items():
                        # rename the bitrates to low, mid, or high. ultra doesn't seem to appear in the adaptive streams
                        name = STREAM_NAMES.get(q, q)
                        streams[name] = s

        return streams

    def _create_api(self):
        """Creates a new CrunchyrollAPI object, initiates its session and
        tries to authenticate it either by using saved credentials or the
        user's username and password.
        """
        if self.options.get("purge_credentials"):
            self.cache.set("device_id", None, expires=0)
            self.cache.set("auth", None, expires=0)

        # use the crunchyroll locale as an override, for backwards compatibility
        locale = self.get_option("locale") or self.session.localization.language_code
        api = CrunchyrollAPI(self.cache,
                             self.session,
                             session_id=self.get_option("session_id"),
                             locale=locale)

        if not self.get_option("session_id"):
            log.debug(f"Creating session with locale: {locale}")
            api.start_session()

            if api.auth:
                log.debug("Using saved credentials")
                login = api.authenticate()
                if login:
                    login_name = login["user"]["username"] or login["user"]["email"]
                    log.info(f"Successfully logged in as '{login_name}'")
            if not api.auth and self.options.get("username"):
                try:
                    log.debug("Attempting to login using username and password")
                    api.login(self.options.get("username"),
                              self.options.get("password"))
                    login = api.authenticate()
                    login_name = login["user"]["username"] or login["user"]["email"]
                    log.info(f"Logged in as '{login_name}'")

                except CrunchyrollAPIError as err:
                    raise PluginError(f"Authentication error: {err.msg}")
            if not api.auth:
                log.warning(
                    "No authentication provided, you won't be able to access "
                    "premium restricted content"
                )

        return api


__plugin__ = Crunchyroll



from unittest.mock import Mock

from freezegun import freeze_time

from streamlink.plugins.vinhlongtv import VinhLongTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVinhLongTV(PluginCanHandleUrl):
    __plugin__ = VinhLongTV

    should_match = [
        "https://www.thvli.vn/live/thvl1-hd",
        "https://www.thvli.vn/live/thvl2-hd",
        "https://www.thvli.vn/live/thvl3-hd",
        "https://www.thvli.vn/live/thvl4-hd",
    ]


@freeze_time("2022-09-25T00:04:45Z")
def test_headers():
    # noinspection PyUnresolvedReferences
    assert VinhLongTV(Mock(), "")._get_headers() == {
        "X-SFD-Date": "20220925000445",
        "X-SFD-Key": "3507c190ae8befda3bfa8e2c00af3c7a",
    }



import logging
import warnings
from inspect import currentframe, getframeinfo
from io import StringIO
from pathlib import Path
from typing import Iterable, Tuple, Type
from unittest.mock import patch

import freezegun
import pytest

from streamlink import logger
from streamlink.exceptions import StreamlinkDeprecationWarning, StreamlinkWarning


@pytest.fixture
def output():
    return StringIO()


@pytest.fixture
def log(request, output: StringIO):
    params = getattr(request, "param", {})
    params.setdefault("format", "[{name}][{levelname}] {message}")
    params.setdefault("style", "{")
    fakeroot = logging.getLogger("streamlink.test")
    with patch("streamlink.logger.root", fakeroot):
        logger.basicConfig(stream=output, **params)
        yield fakeroot
        logger.capturewarnings(False)


class TestLogging:
    @pytest.fixture
    def log_failure(self, request, log: logging.Logger, output: StringIO):
        params = getattr(request, "param", {})
        root = logging.getLogger("streamlink")
        with pytest.raises(Exception) as cm:
            with patch("streamlink.logger.root", root):
                logger.basicConfig(stream=output, **params)
        return cm.value

    @pytest.mark.parametrize("name,level", [
        ("none", logger.NONE),
        ("critical", logger.CRITICAL),
        ("error", logger.ERROR),
        ("warning", logger.WARNING),
        ("info", logger.INFO),
        ("debug", logger.DEBUG),
        ("trace", logger.TRACE),
        ("all", logger.ALL),
    ])
    def test_level_names(self, name: str, level: int):
        assert logging.getLevelName(level) == name
        assert logging.getLevelName(name) == level
        assert logging.getLevelName(name.upper()) == level

    def test_levels_list(self):
        assert logger.levels == ["none", "critical", "error", "warning", "info", "debug", "trace", "all"]

    def test_default_level(self):
        assert logging.getLogger("streamlink").level == logger.WARNING

    def test_level(self, log: logging.Logger, output: StringIO):
        log.setLevel("info")
        log.debug("test")
        output.seek(0)
        assert output.tell() == 0

        log.setLevel("debug")
        log.debug("test")
        assert output.tell() != 0

    def test_level_none(self, log: logging.Logger, output: StringIO):
        log.setLevel("none")
        log.critical("test")
        log.error("test")
        log.warning("test")
        log.info("test")
        log.debug("test")
        log.trace("test")  # type: ignore[attr-defined]
        log.trace("paranoid")  # type: ignore[attr-defined]
        assert not output.getvalue()

    def test_output(self, log: logging.Logger, output: StringIO):
        log.setLevel("debug")
        log.debug("test")
        assert output.getvalue() == "[test][debug] test\n"

    @pytest.mark.parametrize("loglevel,calllevel,expected", [
        (logger.DEBUG, logger.TRACE, ""),
        (logger.TRACE, logger.TRACE, "[test][trace] test\n"),
        (logger.TRACE, logger.DEBUG, "[test][debug] test\n"),
        (logger.TRACE, logger.ALL, ""),
        (logger.ALL, logger.ALL, "[test][all] test\n"),
        (logger.ALL, logger.TRACE, "[test][trace] test\n"),
    ])
    def test_custom_output(self, log: logging.Logger, output: StringIO, loglevel: int, calllevel: int, expected: str):
        log.setLevel(loglevel)
        log.log(calllevel, "test")
        assert output.getvalue() == expected

    # https://github.com/streamlink/streamlink/issues/4862
    @pytest.mark.parametrize("level,levelname", [
        (logger.TRACE, "trace"),
        (logger.ALL, "all"),
    ])
    def test_custom_module_name(self, caplog: pytest.LogCaptureFixture, log: logging.Logger, level: int, levelname: str):
        caplog.set_level(1)
        log = logging.getLogger(self.__class__.__module__)
        getattr(log, levelname)("foo")
        log.log(level, "bar")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("test_logger", levelname, "foo"),
            ("test_logger", levelname, "bar"),
        ]

    @pytest.mark.parametrize("level,expected", [
        (logger.DEBUG, ""),
        (logger.INFO, "[test][info] foo\n[test][info] bar\n"),
    ])
    def test_iter(self, log: logger.StreamlinkLogger, output: StringIO, level: int, expected: str):
        def iterator():
            yield "foo"
            yield "bar"

        log.setLevel(logger.INFO)
        assert list(log.iter(level, iterator())) == ["foo", "bar"]
        assert output.getvalue() == expected

    @pytest.mark.parametrize("log", [
        {"style": "%", "format": "[%(name)s][%(levelname)s] %(message)s"},
    ], indirect=True)
    def test_style_percent(self, log: logging.Logger, output: StringIO):
        log.setLevel("info")
        log.info("test")
        assert output.getvalue() == "[test][info] test\n"

    @pytest.mark.parametrize("log_failure", [{"style": "invalid"}], indirect=True)
    def test_style_invalid(self, log_failure):
        assert type(log_failure) is ValueError
        assert str(log_failure) == "Only {} and % formatting styles are supported"

    @freezegun.freeze_time("2000-01-02T03:04:05.123456Z")
    @pytest.mark.parametrize("log", [
        {"format": "[{asctime}][{name}][{levelname}] {message}"},
    ], indirect=True)
    def test_datefmt_default(self, log: logging.Logger, output: StringIO):
        log.setLevel("info")
        log.info("test")
        assert output.getvalue() == "[03:04:05][test][info] test\n"

    @freezegun.freeze_time("2000-01-02T03:04:05.123456Z")
    @pytest.mark.parametrize("log", [
        {"format": "[{asctime}][{name}][{levelname}] {message}", "datefmt": "%H:%M:%S.%f"},
    ], indirect=True)
    def test_datefmt_custom(self, log: logging.Logger, output: StringIO):
        log.setLevel("info")
        log.info("test")
        assert output.getvalue() == "[03:04:05.123456][test][info] test\n"


class TestCaptureWarnings:
    @staticmethod
    def _warn(messages: Iterable[Tuple[str, Type[Warning]]], filterwarnings=None):
        frame = currentframe()
        assert frame
        assert frame.f_back
        lineno = getframeinfo(frame.f_back).lineno

        with warnings.catch_warnings():
            warnings.filterwarnings(filterwarnings or "always")
            for message, warningclass in messages:
                warnings.warn(message, warningclass, stacklevel=2)

        return lineno

    def test_no_capture(self, log: logging.Logger, output: StringIO):
        with pytest.warns(UserWarning, match=r"^Test warning$"):
            self._warn([("Test warning", UserWarning)])
        assert output.getvalue() == ""

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    @pytest.mark.parametrize("warning,expected,origin", [
        (("Test warning", UserWarning), "[warnings][userwarning] Test warning\n", True),
        (("Test warning", DeprecationWarning), "[warnings][deprecationwarning] Test warning\n", True),
        (("Test warning", FutureWarning), "[warnings][futurewarning] Test warning\n", True),
        (("Test warning", StreamlinkWarning), "[warnings][streamlinkwarning] Test warning\n", False),
        (("Test warning", StreamlinkDeprecationWarning), "[warnings][streamlinkdeprecation] Test warning\n", False),
    ])
    def test_capture(
        self,
        recwarn: pytest.WarningsRecorder,
        log: logging.Logger,
        output: StringIO,
        warning: Tuple[str, Type[Warning]],
        expected: str,
        origin: bool,
    ):
        lineno = self._warn([warning])
        expected += f"  {__file__}:{lineno}\n" if origin else ""
        assert recwarn.list == []
        assert output.getvalue() == expected

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    def test_capture_logrecord(
        self,
        recwarn: pytest.WarningsRecorder,
        caplog: pytest.LogCaptureFixture,
        log: logging.Logger,
    ):
        lineno = self._warn([("Test warning", UserWarning)])
        path = Path(__file__)
        assert recwarn.list == []
        assert [(r.name, r.levelname, r.pathname, r.filename, r.module, r.lineno, r.message) for r in caplog.records] == [
            ("warnings", "userwarning", __file__, path.name, path.stem, lineno, f"Test warning\n  {__file__}:{lineno}"),
        ]

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    def test_capture_consecutive(
        self,
        recwarn: pytest.WarningsRecorder,
        log: logging.Logger,
        output: StringIO,
    ):
        lineno = self._warn([("foo", DeprecationWarning), ("bar", FutureWarning)])
        assert recwarn.list == []
        assert output.getvalue() == (
            f"[warnings][deprecationwarning] foo\n  {__file__}:{lineno}\n"
            + f"[warnings][futurewarning] bar\n  {__file__}:{lineno}\n"
        )

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    def test_capture_consecutive_once(
        self,
        recwarn: pytest.WarningsRecorder,
        log: logging.Logger,
        output: StringIO,
    ):
        lineno = self._warn([("foo", UserWarning), ("foo", UserWarning)], "once")
        assert recwarn.list == []
        assert output.getvalue() == f"[warnings][userwarning] foo\n  {__file__}:{lineno}\n"

    @pytest.mark.parametrize("log", [{"capture_warnings": True}], indirect=["log"])
    @pytest.mark.parametrize("warning", [
        ("Test warning", UserWarning),
        ("Test warning", DeprecationWarning),
        ("Test warning", FutureWarning),
    ])
    def test_ignored(
        self,
        recwarn: pytest.WarningsRecorder,
        log: logging.Logger,
        output: StringIO,
        warning: Tuple[str, Type[Warning]],
    ):
        self._warn([warning], filterwarnings="ignore")
        assert recwarn.list == []
        assert output.getvalue() == ""



from streamlink.plugins.deutschewelle import DeutscheWelle
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDeutscheWelle(PluginCanHandleUrl):
    __plugin__ = DeutscheWelle

    should_match = [
        # Live: en/EN (no channel selection)
        "https://www.dw.com/en/live-tv/s-100825",

        # Live: de/DE (default channel)
        "https://www.dw.com/de/live-tv/s-100817",
        # Live: de/DE (selected channel)
        "https://www.dw.com/de/live-tv/s-100817?channel=5",
        # Live: de/EN (selected channel)
        "https://www.dw.com/de/live-tv/s-100817?channel=1",

        # Live: es/ES (default channel)
        "https://www.dw.com/es/tv-en-vivo/s-100837",
        # Live: es/ES (selected channel)
        "https://www.dw.com/es/tv-en-vivo/s-100837?channel=3",
        # Live: es/DE (selected channel)
        "https://www.dw.com/es/tv-en-vivo/s-100837?channel=5",

        # VOD
        "https://www.dw.com/en/top-stories-in-90-seconds/av-49496622",

        # Audio
        "https://www.dw.com/en/womens-euros-2022-the-end/av-62738085",
    ]



from streamlink.plugins.useetv import UseeTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlUseeTV(PluginCanHandleUrl):
    __plugin__ = UseeTV

    should_match = [
        "http://useetv.com/any",
        "http://useetv.com/any/path",
        "http://www.useetv.com/any",
        "http://www.useetv.com/any/path",
        "https://useetv.com/any",
        "https://useetv.com/any/path",
        "https://www.useetv.com/any",
        "https://www.useetv.com/any/path",
    ]



from streamlink.plugins.pluzz import Pluzz
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPluzz(PluginCanHandleUrl):
    __plugin__ = Pluzz

    should_match = [
        "https://www.france.tv/france-2/direct.html",
        "https://www.france.tv/france-3/direct.html",
        "https://www.france.tv/france-4/direct.html",
        "https://www.france.tv/france-5/direct.html",
        "https://www.france.tv/franceinfo/direct.html",
        "https://www.france.tv/france-2/journal-20h00/141003-edition-du-lundi-8-mai-2017.html",
        "https://france3-regions.francetvinfo.fr/bourgogne-franche-comte/tv/direct/franche-comte",
        "https://www.francetvinfo.fr/en-direct/tv.html",
        "https://www.francetvinfo.fr/meteo/orages/inondations-dans-le-gard-plus-de-deux-mois-de-pluie-en-quelques-heures-des"
        + "-degats-mais-pas-de-victime_4771265.html"
    ]



import unittest

from streamlink.plugins.welt import Welt
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlWelt(PluginCanHandleUrl):
    __plugin__ = Welt

    should_match = [
        "http://welt.de",
        "http://welt.de/",
        "http://welt.de/tv-programm-live-stream/",
        "http://www.welt.de",
        "http://www.welt.de/",
        "http://www.welt.de/tv-programm-live-stream/",
        "https://welt.de",
        "https://welt.de/",
        "https://welt.de/tv-programm-live-stream/",
        "https://www.welt.de",
        "https://www.welt.de/",
        "https://www.welt.de/tv-programm-live-stream/",
    ]


class TestPluginWelt(unittest.TestCase):
    def test_validate_live(self):
        hls_url = Welt._schema.validate("""
            <!DOCTYPE html><html><body>
            <script type="application/json" data-content="VideoPlayer.Config">
                {
                    "title": "foo",
                    "sources": [
                        {
                            "src": "https://foo.bar/baz.mp4?qux",
                            "extension": "mp4"
                        },
                        {
                            "src": "https://foo.bar/baz.m3u8?qux",
                            "extension": "m3u8"
                        }
                    ]
                }
            </script>
            </body></html>
        """)
        self.assertEqual(hls_url, "https://foo.bar/baz.m3u8?qux", "Finds the correct HLS live URL")



# Always get the current version in "editable" installs
# `pip install -e .` / `python setup.py develop`
def _get_version() -> str:
    from pathlib import Path
    from versioningit import get_version
    import streamlink

    return get_version(
        project_dir=Path(streamlink.__file__).parents[2]
    )


# The following _get_version() call will get replaced by versioningit with a static version string when building streamlink
# `pip install .` / `pip wheel .` / `python setup.py build` / `python setup.py bdist_wheel` / etc.
__version__ = "5.2.1"



import asyncio
import errno
import sys
from collections import deque
from pathlib import Path
from threading import Thread
from typing import Callable, Deque, List, Union
from unittest.mock import Mock, patch

import pytest

from streamlink.stream.stream import StreamIO
from streamlink_cli.output import FileOutput, PlayerOutput
from streamlink_cli.streamrunner import PlayerPollThread, StreamRunner, log as streamrunnerlogger
from streamlink_cli.utils.http_server import HTTPServer
from streamlink_cli.utils.progress import Progress
from tests.testutils.handshake import Handshake


TIMEOUT_AWAIT_HANDSHAKE = 1
TIMEOUT_AWAIT_THREADJOIN = 1


class EventedPlayerPollThread(PlayerPollThread):
    POLLING_INTERVAL = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handshake = Handshake()

    def poll(self):
        with self.handshake():
            return super().poll()

    def close(self):
        super().close()
        # Let thread terminate on close()
        self.handshake.go()


class FakeStream(StreamIO):
    """Fake stream implementation, for feeding sample data to the stream runner and simulating read pauses and read errors"""

    def __init__(self) -> None:
        super().__init__()
        self.handshake = Handshake()
        self.data: Deque[Union[bytes, Callable]] = deque()

    # noinspection PyUnusedLocal
    def read(self, *args):
        with self.handshake():
            if not self.data:
                return b""
            data = self.data.popleft()
            return data() if callable(data) else data


class FakeOutput:
    """Common output/http-server/progress interface, for caching all write() calls and simulating write errors"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.handshake = Handshake()
        self.data: List[bytes] = []

    def write(self, data):
        with self.handshake():
            return self._write(data)

    def _write(self, data):
        self.data.append(data)


class FakePlayerOutput(FakeOutput, PlayerOutput):
    pass


class FakeFileOutput(FakeOutput, FileOutput):
    pass


class FakeHTTPServer(FakeOutput, HTTPServer):
    def __init__(self, *args, **kwargs):
        with patch("streamlink_cli.utils.http_server.socket"):
            super().__init__(*args, **kwargs)


class FakeProgress(FakeOutput, Progress):
    # we're not interested in any application logic of the Progress class
    update = print_end = lambda *_, **__: None


class FakeStreamRunner(StreamRunner):
    # override and remove optional typing annotations
    playerpoller: EventedPlayerPollThread
    progress: FakeProgress


@pytest.fixture(autouse=True)
def _logging(caplog: pytest.LogCaptureFixture):
    assert streamrunnerlogger.name == "streamlink.cli"
    caplog.set_level(1, "streamlink")


@pytest.fixture(autouse=True)
def isatty(request: pytest.FixtureRequest):
    with patch("sys.stdout.isatty", return_value=getattr(request, "param", False)):
        yield


@pytest.fixture
def stream():
    stream = FakeStream()
    yield stream
    assert stream.closed


# "stream_runner" fixture dependency declared in downstream scopes
@pytest.fixture
def runnerthread(request: pytest.FixtureRequest, stream_runner: StreamRunner):
    class RunnerThread(Thread):
        exception = None

        def run(self):
            try:
                super().run()
            except BaseException as err:
                self.exception = err

    thread = RunnerThread(
        daemon=True,
        name="Runner thread",
        target=stream_runner.run,
        args=(b"prebuffer",),
    )
    yield thread

    assert_thread_termination(thread, "Runner thread has terminated")

    exception = getattr(request, "param", {}).get("exception", None)
    assert isinstance(thread.exception, type(exception))
    assert str(thread.exception) == str(exception)


async def assert_handshake_steps(*items):
    """
    Run handshake steps concurrently, to not be dependent too much on implementation details and the order of handshakes.
    For example, concurrently await one read(), one write() and one progress() call.
    """
    steps = asyncio.gather(
        *(item.handshake.asyncstep(TIMEOUT_AWAIT_HANDSHAKE) for item in items),
        return_exceptions=True,
    )
    assert await steps == [True for _ in items]


def assert_thread_termination(thread: Thread, assertion: str):
    thread.join(TIMEOUT_AWAIT_THREADJOIN)
    assert not thread.is_alive(), assertion


class TestPlayerOutput:
    @pytest.fixture
    def player_process(self):
        player_process = Mock()
        player_process.poll = Mock(return_value=None)

        return player_process

    @pytest.fixture
    def output(self, player_process: Mock):
        with patch("subprocess.Popen") as mock_popen, \
             patch("streamlink_cli.output.sleep"):
            mock_popen.return_value = player_process
            output = FakePlayerOutput("mocked")
            output.open()
            yield output

    @pytest.fixture
    def stream_runner(self, stream: FakeStream, output: FakePlayerOutput):
        with patch("streamlink_cli.streamrunner.PlayerPollThread", EventedPlayerPollThread):
            stream_runner = StreamRunner(stream, output)
            assert isinstance(stream_runner.playerpoller, EventedPlayerPollThread)
            assert not stream_runner.playerpoller.is_alive()
            assert not stream_runner.is_http
            assert not stream_runner.progress
            yield stream_runner
            assert not stream_runner.playerpoller.is_alive()

    @pytest.mark.asyncio
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.asyncio
    async def test_paused(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        delayed = Handshake()

        def item():
            with delayed():
                return b"delayed"

        stream.data.append(item)

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]
        assert not delayed.wait_ready(0), "Delayed chunk has not been read yet"

        # attempt reading delayed chunk
        stream.handshake.go()
        assert delayed.wait_ready(TIMEOUT_AWAIT_HANDSHAKE), "read() call of delayed chunk is paused"
        assert output.data == [b"prebuffer"]

        assert not stream.closed, "Stream is not closed"

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # unpause delayed chunk
        delayed.go()
        assert stream.handshake.wait_done(TIMEOUT_AWAIT_HANDSHAKE), "Delayed chunk has successfully been read"
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer", b"delayed"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"delayed"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "writeerror,runnerthread",
        [
            pytest.param(
                OSError(errno.EPIPE, "Broken pipe"),
                {},
                id="Acceptable error: EPIPE",
            ),
            pytest.param(
                OSError(errno.EINVAL, "Invalid argument"),
                {},
                id="Acceptable error: EINVAL",
            ),
            pytest.param(
                OSError(errno.ECONNRESET, "Connection reset"),
                {},
                id="Acceptable error: ECONNRESET",
            ),
            pytest.param(
                OSError("Unknown error"),
                {"exception": OSError("Error when writing to output: Unknown error, exiting")},
                id="Non-acceptable error",
            ),
        ],
        indirect=["runnerthread"],
    )
    async def test_player_close(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        player_process: Mock,
        writeerror: Exception,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        assert not stream.closed, "Stream is not closed yet"

        # close player
        with patch.object(output, "_write", side_effect=writeerror):
            # let player process terminate with code 0 and poll process once
            player_process.poll.return_value = 0
            await assert_handshake_steps(stream_runner.playerpoller)
            assert_thread_termination(stream_runner.playerpoller, "Polling has stopped after player process terminated")

            assert stream.closed, "Stream got closed after the player was closed"

            # read and write next chunk (write will now also raise)
            await assert_handshake_steps(stream, output)
            assert output.data == [b"prebuffer", b"foo"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Player closed"),
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.asyncio
    async def test_player_close_paused(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        player_process: Mock,
    ):
        delayed = Handshake()

        def item():
            with delayed():
                return b""

        stream.data.append(item)

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]
        assert not delayed.wait_ready(0), "Delayed chunk has not been read yet"

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        stream.handshake.go()
        assert delayed.wait_ready(TIMEOUT_AWAIT_HANDSHAKE), "read() call of delayed chunk is paused"
        assert output.data == [b"prebuffer"]

        assert not stream.closed, "Stream is not closed yet"

        # let player process terminate with code 0 and poll process once
        player_process.poll.return_value = 0
        await assert_handshake_steps(stream_runner.playerpoller)
        assert_thread_termination(stream_runner.playerpoller, "Polling has stopped after player process terminated")

        assert stream.closed, "Stream got closed after the player was closed, even if the stream was paused"

        # unpause delayed chunk (stream end)
        delayed.go()
        assert stream.handshake.wait_done(TIMEOUT_AWAIT_HANDSHAKE), "Delayed chunk has successfully been read"
        assert output.data == [b"prebuffer"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Player closed"),
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "runnerthread",
        [{"exception": OSError("Error when reading from stream: Read timeout, exiting")}],
        indirect=["runnerthread"],
    )
    async def test_readerror(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
    ):
        # make next read() call raise a read-timeout error
        stream.data.append(Mock(side_effect=OSError("Read timeout")))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # poll player process
        await assert_handshake_steps(stream_runner.playerpoller)
        assert stream_runner.playerpoller.is_alive()

        # read stream (will raise a read timeout)
        await assert_handshake_steps(stream)

        # poll player process again
        await assert_handshake_steps(stream_runner.playerpoller)
        assert_thread_termination(stream_runner.playerpoller, "Polling has stopped on read error")

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]


class TestHTTPServer:
    @pytest.fixture
    def output(self):
        return FakeHTTPServer()

    @pytest.fixture
    def stream_runner(self, stream: FakeStream, output: FakeHTTPServer):
        stream_runner = StreamRunner(stream, output)
        assert not stream_runner.playerpoller
        assert not stream_runner.progress
        assert stream_runner.is_http
        yield stream_runner

    @pytest.mark.asyncio
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakeHTTPServer,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output)
        assert output.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo"]

        # read and write next chunk
        await assert_handshake_steps(stream, output)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]

    @pytest.mark.parametrize(
        "writeerror,logs,runnerthread",
        [
            pytest.param(
                OSError(errno.EPIPE, "Broken pipe"),
                True,
                {},
                id="Acceptable error: EPIPE",
            ),
            pytest.param(
                OSError(errno.EINVAL, "Invalid argument"),
                True,
                {},
                id="Acceptable error: EINVAL",
            ),
            pytest.param(
                OSError(errno.ECONNRESET, "Connection reset"),
                True,
                {},
                id="Acceptable error: ECONNRESET",
            ),
            pytest.param(
                OSError("Unknown error"),
                False,
                {"exception": OSError("Error when writing to output: Unknown error, exiting")},
                id="Non-acceptable error",
            ),
        ],
        indirect=["runnerthread"],
    )
    def test_writeerror(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakePlayerOutput,
        logs: bool,
        writeerror: Exception,
    ):
        runnerthread.start()

        with patch.object(output, "_write", side_effect=writeerror):
            assert output.handshake.step(TIMEOUT_AWAIT_HANDSHAKE)
            assert output.data == []

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        expectedlogs = (
            ([("streamrunner", "info", "HTTP connection closed")] if logs else [])
            + [("streamrunner", "info", "Stream ended")]
        )
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == expectedlogs


@pytest.mark.parametrize(
    "isatty,force_progress",
    [
        pytest.param(False, True, id="No TTY, force"),
        pytest.param(True, False, id="TTY, no force"),
    ],
    indirect=["isatty"],
)
class TestHasProgress:
    @pytest.mark.parametrize(
        "output",
        [
            pytest.param(
                FakePlayerOutput("mocked"),
                id="Player output without record",
            ),
            pytest.param(
                FakeFileOutput(fd=Mock()),
                id="FileOutput with file descriptor",
            ),
            pytest.param(
                FakeHTTPServer(),
                id="HTTPServer",
            ),
        ],
    )
    def test_no_progress(
        self,
        output: Union[FakePlayerOutput, FakeFileOutput, FakeHTTPServer],
        isatty: bool,
        force_progress: bool,
    ):
        stream_runner = FakeStreamRunner(StreamIO(), output, force_progress)
        assert not stream_runner.progress

    @pytest.mark.parametrize(
        "output,expected",
        [
            pytest.param(
                FakePlayerOutput("mocked", record=FakeFileOutput(Path("record"))),
                Path("record"),
                id="PlayerOutput with record",
            ),
            pytest.param(
                FakeFileOutput(filename=Path("filename")),
                Path("filename"),
                id="FileOutput with file name",
            ),
            pytest.param(
                FakeFileOutput(record=FakeFileOutput(filename=Path("record"))),
                Path("record"),
                id="FileOutput with record",
            ),
            pytest.param(
                FakeFileOutput(filename=Path("filename"), record=FakeFileOutput(filename=Path("record"))),
                Path("filename"),
                id="FileOutput with file name and record",
            ),
        ],
    )
    def test_has_progress(
        self,
        output: Union[FakePlayerOutput, FakeFileOutput],
        isatty: bool,
        force_progress: bool,
        expected: Path,
    ):
        stream_runner = FakeStreamRunner(StreamIO(), output, force_progress)
        assert stream_runner.progress
        assert not stream_runner.progress.is_alive()
        assert stream_runner.progress.stream is sys.stderr
        assert stream_runner.progress.path == expected


class TestProgress:
    @pytest.fixture
    def output(self):
        yield FakeFileOutput(Path("filename"))

    @pytest.fixture
    def stream_runner(self, stream: FakeStream, output: FakeFileOutput):
        with patch("streamlink_cli.streamrunner.Progress", FakeProgress):
            stream_runner = FakeStreamRunner(stream, output, True)
            assert not stream_runner.playerpoller
            assert not stream_runner.is_http
            assert isinstance(stream_runner.progress, FakeProgress)
            assert stream_runner.progress.path == Path("filename")
            assert not stream_runner.progress.is_alive()
            yield stream_runner
            assert not stream_runner.progress.is_alive()

    @pytest.mark.asyncio
    async def test_read_write(
        self,
        caplog: pytest.LogCaptureFixture,
        runnerthread: Thread,
        stream_runner: FakeStreamRunner,
        stream: FakeStream,
        output: FakeFileOutput,
    ):
        stream.data.extend((b"foo", b"bar"))

        runnerthread.start()
        assert output.data == []

        # write prebuffer
        await assert_handshake_steps(output, stream_runner.progress)
        assert output.data == [b"prebuffer"]
        assert stream_runner.progress.data == [b"prebuffer"]

        # read and write next chunk
        await assert_handshake_steps(stream, output, stream_runner.progress)
        assert output.data == [b"prebuffer", b"foo"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo"]

        # read and write next chunk
        await assert_handshake_steps(stream, output, stream_runner.progress)
        assert output.data == [b"prebuffer", b"foo", b"bar"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo", b"bar"]

        assert not stream.closed, "Stream is not closed"

        # read stream end
        await assert_handshake_steps(stream)
        assert output.data == [b"prebuffer", b"foo", b"bar"]
        assert stream_runner.progress.data == [b"prebuffer", b"foo", b"bar"]

        # wait for runner thread to terminate first before asserting log records
        assert_thread_termination(runnerthread, "Runner thread has terminated")
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("streamrunner", "info", "Stream ended"),
        ]



from streamlink.plugins.delfi import Delfi
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDelfi(PluginCanHandleUrl):
    __plugin__ = Delfi

    should_match = [
        # delfi.lt live (YouTube)
        "https://www.delfi.lt/video/tv/",
        # delfi.lt VOD
        "https://www.delfi.lt/video/laidos/dakaras/dakaras-2022-koki-tiksla-turi-pirmasis-lietuviskas-sunkvezimio-ekipazas.d"
        + "?id=89058633",

        # delfi.lv VOD
        "http://www.delfi.lv/delfi-tv-ar-jani-domburu/pilnie-raidijumi/delfi-tv-ar-jani-domburu-atbild"
        + "-veselibas-ministre-anda-caksa-pilna-intervija?id=49515013",
        # delfi.lv VOD (YouTube)
        "https://www.delfi.lv/news/national/politics/video-gads-ko-varetu-aizmirst-bet-nesanak-spilgtako-notikumu-atskats.d"
        + "?id=53912761",

        # delfi.ee live
        "https://sport.delfi.ee/artikkel/95517317/otse-delfi-tv-s-kalevcramo-voitis-tartu-ulikooli-vastu-avapoolaja-14"
        + "-punktiga",
    ]



from streamlink.plugins.mixcloud import Mixcloud
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMixcloud(PluginCanHandleUrl):
    __plugin__ = Mixcloud

    should_match_groups = [
        ("http://mixcloud.com/live/user", {"user": "user"}),
        ("http://www.mixcloud.com/live/user", {"user": "user"}),
        ("https://mixcloud.com/live/user", {"user": "user"}),
        ("https://www.mixcloud.com/live/user", {"user": "user"}),
        ("https://www.mixcloud.com/live/user/anything", {"user": "user"}),
    ]

    should_not_match = [
        "https://www.mixcloud.com/",
        "https://www.mixcloud.com/live",
        "https://www.mixcloud.com/live/",
        "https://www.mixcloud.com/other",
    ]



"""
$description Spanish live TV channels from Atresmedia Television, including Antena 3 and laSexta.
$url atresplayer.com
$type live
$region Spain
"""

import logging
import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?atresplayer\.com/"
))
class AtresPlayer(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = update_scheme("https://", f"{self.url.rstrip('/')}/")

    def _get_streams(self):
        path = urlparse(self.url).path
        api_url = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"""window.__PRELOADED_STATE__\s*=\s*({.*?});""", re.DOTALL),
            validate.none_or_all(
                validate.get(1),
                validate.parse_json(),
                {"links": {path: {"href": validate.url()}}},
                validate.get(("links", path, "href")),
            ),
        ))
        if not api_url:
            return
        log.debug(f"API URL: {api_url}")

        player_api_url = self.session.http.get(api_url, schema=validate.Schema(
            validate.parse_json(),
            {"urlVideo": validate.url()},
            validate.get("urlVideo"),
        ))

        log.debug(f"Player API URL: {player_api_url}")
        sources = self.session.http.get(player_api_url, acceptable_status=(200, 403), schema=validate.Schema(
            validate.parse_json(),
            validate.any(
                {
                    "error": str,
                    "error_description": str,
                },
                {
                    "sources": [
                        validate.all(
                            {
                                "src": validate.url(),
                                validate.optional("type"): str,
                            },
                            validate.union_get("type", "src"),
                        ),
                    ],
                },
            ),
        ))
        if "error" in sources:
            log.error(f"Player API error: {sources['error']} - {sources['error_description']}")
            return

        for streamtype, streamsrc in sources.get("sources"):
            log.debug(f"Stream source: {streamsrc} ({streamtype or 'n/a'})")

            if streamtype == "application/vnd.apple.mpegurl":
                streams = HLSStream.parse_variant_playlist(self.session, streamsrc)
                if not streams:
                    yield "live", HLSStream(self.session, streamsrc)
                else:
                    yield from streams.items()
            elif streamtype == "application/dash+xml":
                yield from DASHStream.parse_manifest(self.session, streamsrc).items()


__plugin__ = AtresPlayer



from streamlink.plugins.tvp import TVP
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTVP(PluginCanHandleUrl):
    __plugin__ = TVP

    should_match_groups = [
        # live
        ("https://stream.tvp.pl", {}),
        ("https://stream.tvp.pl/", {}),
        ("https://stream.tvp.pl/?channel_id=63759349", {"channel_id": "63759349"}),
        ("https://stream.tvp.pl/?channel_id=14812849", {"channel_id": "14812849"}),
        # old live URLs
        ("https://tvpstream.vod.tvp.pl", {}),
        ("https://tvpstream.vod.tvp.pl/", {}),
        ("https://tvpstream.vod.tvp.pl/?channel_id=63759349", {"channel_id": "63759349"}),
        ("https://tvpstream.vod.tvp.pl/?channel_id=14812849", {"channel_id": "14812849"}),

        # VOD
        (
            "https://vod.tvp.pl/filmy-dokumentalne,163/krolowa-wladczyni-i-matka,284734",
            {"vod_id": "284734"},
        ),
        # VOD episode
        (
            "https://vod.tvp.pl/programy,88/z-davidem-attenborough-dokola-swiata-odcinki,284703/odcinek-2,S01E02,319220",
            {"vod_id": "319220"},
        ),

        # tvp.info
        (("tvp_info", "https://tvp.info/"), {}),
        (("tvp_info", "https://www.tvp.info/"), {}),
        (("tvp_info", "https://www.tvp.info/65275202/13012023-0823"), {}),
    ]

    should_not_match = [
        "https://tvp.pl/",
        "https://vod.tvp.pl/",
    ]



from streamlink.plugins.onetv import OneTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlOneTV(PluginCanHandleUrl):
    __plugin__ = OneTV

    should_match = [
        "https://www.1tv.ru/live",
        "http://www.1tv.ru/live",
        "https://static.1tv.ru/eump/embeds/1tv_live_orbit-plus-4.html?muted=no",
        "https://static.1tv.ru/eump/pages/1tv_live.html",
        "https://static.1tv.ru/eump/pages/1tv_live_orbit-plus-4.html",
    ]

    should_not_match = [
        "http://www.1tv.ru/some-show/some-programme-2018-03-10",
        "https://www.ctc.ru/online",
        "http://www.ctc.ru/online",
        "https://www.chetv.ru/online",
        "http://www.chetv.ru/online",
        "https://www.ctclove.ru/online",
        "http://www.ctclove.ru/online",
        "https://www.domashny.ru/online",
        "http://www.domashny.ru/online",
    ]



import unittest

from streamlink.utils.times import hours_minutes_seconds, seconds_to_hhmmss


class TestUtilsTimes(unittest.TestCase):
    def test_hours_minutes_seconds(self):
        self.assertEqual(hours_minutes_seconds("00:01:30"), 90)
        self.assertEqual(hours_minutes_seconds("01:20:15"), 4815)
        self.assertEqual(hours_minutes_seconds("26:00:00"), 93600)

        self.assertEqual(hours_minutes_seconds("07"), 7)
        self.assertEqual(hours_minutes_seconds("444"), 444)
        self.assertEqual(hours_minutes_seconds("8888"), 8888)

        self.assertEqual(hours_minutes_seconds("01h"), 3600)
        self.assertEqual(hours_minutes_seconds("01h22m33s"), 4953)
        self.assertEqual(hours_minutes_seconds("01H22M37S"), 4957)
        self.assertEqual(hours_minutes_seconds("01h30s"), 3630)
        self.assertEqual(hours_minutes_seconds("1m33s"), 93)
        self.assertEqual(hours_minutes_seconds("55s"), 55)

        self.assertEqual(hours_minutes_seconds("-00:01:40"), 100)
        self.assertEqual(hours_minutes_seconds("-00h02m30s"), 150)

        self.assertEqual(hours_minutes_seconds("02:04"), 124)
        self.assertEqual(hours_minutes_seconds("1:10"), 70)
        self.assertEqual(hours_minutes_seconds("10:00"), 600)

        with self.assertRaises(ValueError):
            hours_minutes_seconds("FOO")

        with self.assertRaises(ValueError):
            hours_minutes_seconds("BAR")

        with self.assertRaises(ValueError):
            hours_minutes_seconds("11:ERR:00")

    def test_seconds_to_hhmmss(self):
        self.assertEqual(seconds_to_hhmmss(0), "00:00:00")
        self.assertEqual(seconds_to_hhmmss(1), "00:00:01")
        self.assertEqual(seconds_to_hhmmss(60), "00:01:00")
        self.assertEqual(seconds_to_hhmmss(3600), "01:00:00")

        self.assertEqual(seconds_to_hhmmss(13997), "03:53:17")
        self.assertEqual(seconds_to_hhmmss(13997.4), "03:53:17.4")



from streamlink.plugins.huajiao import Huajiao
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlHuajiao(PluginCanHandleUrl):
    __plugin__ = Huajiao

    should_match = [
        "http://www.huajiao.com/l/123123123",
        "https://www.huajiao.com/l/123123123",
    ]



"""
$description Japanese live-streaming and video hosting social platform.
$url live.nicovideo.jp
$type live, vod
$account Required by some streams
$notes Timeshift is supported
"""

import logging
import re
from threading import Event
from urllib.parse import urljoin

from streamlink.plugin import Plugin, PluginError, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.plugin.api.websocket import WebsocketClient
from streamlink.stream.hls import HLSStream, HLSStreamReader
from streamlink.utils.parse import parse_json
from streamlink.utils.times import hours_minutes_seconds
from streamlink.utils.url import update_qsd

log = logging.getLogger(__name__)


class NicoLiveWsClient(WebsocketClient):
    STREAM_OPENED_TIMEOUT = 6

    ready: Event
    opened: Event
    hls_stream_url: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.opened = Event()
        self.ready = Event()

    def on_open(self, wsapp):
        super().on_open(wsapp)
        self.send_playerversion()
        self.send_getpermit()

    def on_message(self, wsapp, data: str):
        log.debug(f"Received: {data}")
        message = parse_json(data)
        msgtype = message.get("type")
        msgdata = message.get("data", {})

        if msgtype == "ping":
            self.send_pong()

        elif msgtype == "stream" and msgdata.get("protocol") == "hls" and msgdata.get("uri"):
            self.hls_stream_url = msgdata.get("uri")
            self.ready.set()
            if self.opened.wait(self.STREAM_OPENED_TIMEOUT):
                log.debug("Stream opened, keeping websocket connection alive")
            else:
                log.info("Closing websocket connection")
                self.close()

        elif msgtype == "disconnect":
            reason = msgdata.get("reason", "Unknown reason")
            log.info(f"Received disconnect message: {reason}")
            self.close()

    def send_playerversion(self):
        self.send_json({
            "type": "startWatching",
            "data": {
                "stream": {
                    "quality": "abr",
                    "protocol": "hls",
                    "latency": "high",
                    "chasePlay": False
                },
                "room": {
                    "protocol": "webSocket",
                    "commentable": True
                },
                "reconnect": False
            }
        })

    def send_getpermit(self):
        self.send_json({
            "type": "getAkashic",
            "data": {
                "chasePlay": False
            }
        })

    def send_pong(self):
        self.send_json({"type": "pong"})
        self.send_json({"type": "keepSeat"})


class NicoLiveHLSStreamReader(HLSStreamReader):
    stream: "NicoLiveHLSStream"

    def open(self):
        self.stream.wsclient.opened.set()
        super().open()

    def close(self):
        super().close()
        self.stream.wsclient.close()


class NicoLiveHLSStream(HLSStream):
    __reader__ = NicoLiveHLSStreamReader
    wsclient: NicoLiveWsClient

    def set_wsclient(self, wsclient: NicoLiveWsClient):
        self.wsclient = wsclient


@pluginmatcher(re.compile(
    r"https?://(?P<domain>live\d*\.nicovideo\.jp)/watch/(lv|co)\d+"
))
@pluginargument(
    "email",
    sensitive=True,
    argument_name="niconico-email",
    metavar="EMAIL",
    help="The email or phone number associated with your Niconico account",
)
@pluginargument(
    "password",
    sensitive=True,
    argument_name="niconico-password",
    metavar="PASSWORD",
    help="The password of your Niconico account",
)
@pluginargument(
    "user-session",
    sensitive=True,
    argument_name="niconico-user-session",
    metavar="VALUE",
    help="""
        Value of the user-session token.

        Can be used as an alternative to providing a password.
    """,
)
@pluginargument(
    "purge-credentials",
    argument_name="niconico-purge-credentials",
    action="store_true",
    help="Purge cached Niconico credentials to initiate a new session and reauthenticate.",
)
@pluginargument(
    "timeshift-offset",
    type=hours_minutes_seconds,
    argument_name="niconico-timeshift-offset",
    metavar="[HH:]MM:SS",
    default=None,
    help="""
        Amount of time to skip from the beginning of a stream.

        Default is 00:00:00.
    """,
)
class NicoLive(Plugin):
    STREAM_READY_TIMEOUT = 6
    LOGIN_URL = "https://account.nicovideo.jp/login/redirector"
    LOGIN_URL_PARAMS = {
        "show_button_twitter": 1,
        "show_button_facebook": 1,
        "next_url": "/",
    }

    wsclient: NicoLiveWsClient

    def _get_streams(self):
        if self.get_option("purge_credentials"):
            self.clear_cookies()
            log.info("All credentials were successfully removed")

        self.session.http.headers.update({
            "User-Agent": useragents.CHROME,
        })

        self.niconico_web_login()

        wss_api_url = self.get_wss_api_url()
        if not wss_api_url:
            log.error(
                "Failed to get wss_api_url. "
                "Please check if the URL is correct, "
                "and make sure your account has access to the video."
            )
            return

        self.wsclient = NicoLiveWsClient(self.session, wss_api_url)
        self.wsclient.start()

        hls_stream_url = self._get_hls_stream_url()
        if not hls_stream_url:
            return

        offset = self.get_option("timeshift-offset")
        if offset and "timeshift" in wss_api_url:
            hls_stream_url = update_qsd(hls_stream_url, {"start": offset})

        for quality, stream in NicoLiveHLSStream.parse_variant_playlist(self.session, hls_stream_url).items():
            stream.set_wsclient(self.wsclient)
            yield quality, stream

    def _get_hls_stream_url(self):
        log.debug(f"Waiting for permit (for at most {self.STREAM_READY_TIMEOUT} seconds)...")
        if not self.wsclient.ready.wait(self.STREAM_READY_TIMEOUT) or not self.wsclient.is_alive():
            log.error("Waiting for permit timed out.")
            self.wsclient.close()
            return

        return self.wsclient.hls_stream_url

    def get_wss_api_url(self):
        try:
            data = self.session.http.get(self.url, schema=validate.Schema(
                validate.parse_html(),
                validate.xml_find(".//script[@id='embedded-data'][@data-props]"),
                validate.get("data-props"),
                validate.parse_json(),
                {"site": {
                    "relive": {
                        "webSocketUrl": validate.url(scheme="wss")
                    },
                    validate.optional("frontendId"): int
                }},
                validate.get("site"),
                validate.union_get(("relive", "webSocketUrl"), "frontendId")
            ))
        except PluginError:
            return

        wss_api_url, frontend_id = data
        if frontend_id is not None:
            wss_api_url = update_qsd(wss_api_url, {"frontend_id": frontend_id})

        return wss_api_url

    def niconico_web_login(self):
        user_session = self.get_option("user-session")
        email = self.get_option("email")
        password = self.get_option("password")

        if user_session is not None:
            log.info("Logging in via provided user session cookie")
            self.session.http.cookies.set(
                "user_session",
                user_session,
                path="/",
                domain="nicovideo.jp"
            )
            self.save_cookies()

        elif self.session.http.cookies.get("user_session"):
            log.info("Logging in via cached user session cookie")

        elif email is not None and password is not None:
            log.info("Logging in via provided email and password")
            root = self.session.http.post(
                self.LOGIN_URL,
                data={"mail_tel": email, "password": password},
                params=self.LOGIN_URL_PARAMS,
                schema=validate.Schema(validate.parse_html()),
            )

            if self.session.http.cookies.get("user_session"):
                log.info("Logged in.")
                self.save_cookies()
                return

            input_with_value = {}
            for elem in root.xpath(".//form[@action]//input"):
                if elem.attrib.get("value"):
                    input_with_value[elem.attrib.get("name")] = elem.attrib.get("value")
                elif elem.attrib.get("id") == "oneTimePw":
                    maxlength = int(elem.attrib.get("maxlength"))
                    oneTimePw = self.input_ask("Enter the 6 digit number included in email")
                    if len(oneTimePw) > maxlength:
                        log.error("invalid user input")
                        return
                    input_with_value[elem.attrib.get("name")] = oneTimePw
                else:
                    log.debug(f"unknown input: {elem.attrib.get('name')}")

            root = self.session.http.post(
                urljoin("https://account.nicovideo.jp", root.xpath("string(.//form[@action]/@action)")),
                data=input_with_value,
                schema=validate.Schema(validate.parse_html()),
            )
            log.debug(f"Cookies: {self.session.http.cookies.get_dict()}")
            if self.session.http.cookies.get("user_session") is None:
                error = root.xpath("string(//div[@class='formError']/div/text())")
                log.warning(f"Login failed: {error or 'unknown reason'}")
            else:
                log.info("Logged in.")
                self.save_cookies()


__plugin__ = NicoLive



import ssl
import time
from typing import Any, Dict, Pattern, Tuple

import requests.adapters
import urllib3
from requests import PreparedRequest, Request, Session

from streamlink.exceptions import PluginError
from streamlink.packages.requests_file import FileAdapter
from streamlink.plugin.api import useragents
from streamlink.utils.parse import parse_json, parse_xml


class _HTTPResponse(urllib3.response.HTTPResponse):
    def __init__(self, *args, **kwargs):
        # Always enforce content length validation!
        # This fixes a bug in requests which doesn't raise errors on HTTP responses where
        # the "Content-Length" header doesn't match the response's body length.
        # https://github.com/psf/requests/issues/4956#issuecomment-573325001
        #
        # Summary:
        # This bug is related to urllib3.response.HTTPResponse.stream() which calls urllib3.response.HTTPResponse.read() as
        # a wrapper for http.client.HTTPResponse.read(amt=...), where no http.client.IncompleteRead exception gets raised
        # due to "backwards compatiblity" of an old bug if a specific amount is attempted to be read on an incomplete response.
        #
        # urllib3.response.HTTPResponse.read() however has an additional check implemented via the enforce_content_length
        # parameter, but it doesn't check by default and requests doesn't set the parameter for enabling it either.
        #
        # Fix this by overriding urllib3.response.HTTPResponse's constructor and always setting enforce_content_length to True,
        # as there is no way to make requests set this parameter on its own.
        kwargs["enforce_content_length"] = True
        super().__init__(*args, **kwargs)


# override all urllib3.response.HTTPResponse references in requests.adapters.HTTPAdapter.send
urllib3.connectionpool.HTTPConnectionPool.ResponseCls = _HTTPResponse  # type: ignore[attr-defined]
requests.adapters.HTTPResponse = _HTTPResponse  # type: ignore[misc]


# Never convert percent-encoded characters to uppercase in urllib3>=1.25.8.
# This is required for sites which compare request URLs byte by byte and return different responses depending on that.
# Older versions of urllib3 are not compatible with this override and will always convert to uppercase characters.
#
# https://datatracker.ietf.org/doc/html/rfc3986#section-2.1
# > The uppercase hexadecimal digits 'A' through 'F' are equivalent to
# > the lowercase digits 'a' through 'f', respectively.  If two URIs
# > differ only in the case of hexadecimal digits used in percent-encoded
# > octets, they are equivalent.  For consistency, URI producers and
# > normalizers should use uppercase hexadecimal digits for all percent-
# > encodings.
class Urllib3UtilUrlPercentReOverride:
    _re_percent_encoding: Pattern = urllib3.util.url.PERCENT_RE  # type: ignore[attr-defined]

    # urllib3>=1.25.8
    # https://github.com/urllib3/urllib3/blame/1.25.8/src/urllib3/util/url.py#L219-L227
    @classmethod
    def subn(cls, repl: Any, string: str, count: Any = None) -> Tuple[str, int]:
        return string, len(cls._re_percent_encoding.findall(string))


urllib3.util.url.PERCENT_RE = Urllib3UtilUrlPercentReOverride  # type: ignore[attr-defined]


# requests.Request.__init__ keywords, except for "hooks"
_VALID_REQUEST_ARGS = "method", "url", "headers", "files", "data", "params", "auth", "cookies", "json"


class HTTPSession(Session):
    params: Dict

    def __init__(self):
        super().__init__()

        self.headers['User-Agent'] = useragents.FIREFOX
        self.timeout = 20.0

        self.mount('file://', FileAdapter())

    @classmethod
    def determine_json_encoding(cls, sample):
        """
        Determine which Unicode encoding the JSON text sample is encoded with

        RFC4627 (http://www.ietf.org/rfc/rfc4627.txt) suggests that the encoding of JSON text can be determined
        by checking the pattern of NULL bytes in first 4 octets of the text.
        :param sample: a sample of at least 4 bytes of the JSON text
        :return: the most likely encoding of the JSON text
        """
        nulls_at = [i for i, j in enumerate(bytearray(sample[:4])) if j == 0]
        if nulls_at == [0, 1, 2]:
            return "UTF-32BE"
        elif nulls_at == [0, 2]:
            return "UTF-16BE"
        elif nulls_at == [1, 2, 3]:
            return "UTF-32LE"
        elif nulls_at == [1, 3]:
            return "UTF-16LE"
        else:
            return "UTF-8"

    @classmethod
    def json(cls, res, *args, **kwargs):
        """Parses JSON from a response."""
        # if an encoding is already set then use the provided encoding
        if res.encoding is None:
            res.encoding = cls.determine_json_encoding(res.content[:4])
        return parse_json(res.text, *args, **kwargs)

    @classmethod
    def xml(cls, res, *args, **kwargs):
        """Parses XML from a response."""
        return parse_xml(res.text, *args, **kwargs)

    def resolve_url(self, url):
        """Resolves any redirects and returns the final URL."""
        return self.get(url, stream=True).url

    @staticmethod
    def valid_request_args(**req_keywords) -> Dict:
        return {k: v for k, v in req_keywords.items() if k in _VALID_REQUEST_ARGS}

    def prepare_new_request(self, **req_keywords) -> PreparedRequest:
        valid_args = self.valid_request_args(**req_keywords)
        valid_args.setdefault("method", "GET")
        request = Request(**valid_args)

        # prepare request with the session context, which might add params, headers, cookies, etc.
        return self.prepare_request(request)

    def request(self, method, url, *args, **kwargs):
        acceptable_status = kwargs.pop("acceptable_status", [])
        exception = kwargs.pop("exception", PluginError)
        headers = kwargs.pop("headers", {})
        params = kwargs.pop("params", {})
        proxies = kwargs.pop("proxies", self.proxies)
        raise_for_status = kwargs.pop("raise_for_status", True)
        schema = kwargs.pop("schema", None)
        session = kwargs.pop("session", None)
        timeout = kwargs.pop("timeout", self.timeout)
        total_retries = kwargs.pop("retries", 0)
        retry_backoff = kwargs.pop("retry_backoff", 0.3)
        retry_max_backoff = kwargs.pop("retry_max_backoff", 10.0)
        retries = 0

        if session:
            headers.update(session.headers)
            params.update(session.params)

        while True:
            try:
                res = super().request(
                    method,
                    url,
                    headers=headers,
                    params=params,
                    timeout=timeout,
                    proxies=proxies,
                    *args,
                    **kwargs
                )
                if raise_for_status and res.status_code not in acceptable_status:
                    res.raise_for_status()
                break
            except KeyboardInterrupt:
                raise
            except Exception as rerr:
                if retries >= total_retries:
                    err = exception(f"Unable to open URL: {url} ({rerr})")
                    err.err = rerr
                    raise err
                retries += 1
                # back off retrying, but only to a maximum sleep time
                delay = min(retry_max_backoff,
                            retry_backoff * (2 ** (retries - 1)))
                time.sleep(delay)

        if schema:
            res = schema.validate(res.text, name="response text", exception=PluginError)

        return res


class TLSSecLevel1Adapter(requests.adapters.HTTPAdapter):
    def init_poolmanager(self, *args, **kwargs):
        ctx = ssl.create_default_context()
        ctx.set_ciphers("DEFAULT:@SECLEVEL=1")
        kwargs["ssl_context"] = ctx
        return super().init_poolmanager(*args, **kwargs)


__all__ = ["HTTPSession", "TLSSecLevel1Adapter"]



from streamlink.plugins.tv999 import TV999
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTV999(PluginCanHandleUrl):
    __plugin__ = TV999

    should_match = [
        "http://tv999.bg/live.html",
        "http://www.tv999.bg/live.html",
        "https://tv999.bg/live",
        "https://tv999.bg/live.html",
        "https://www.tv999.bg/live.html",
    ]

    should_not_match = [
        "http://tv999.bg/",
    ]



from streamlink.plugins.tviplayer import TVIPlayer
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTVIPlayer(PluginCanHandleUrl):
    __plugin__ = TVIPlayer

    should_match = [
        "https://tviplayer.iol.pt/direto/TVI24",
        "https://tviplayer.iol.pt/direto/TVI_AFRICA",
        "https://tviplayer.iol.pt/programa/prisioneira/5c890ae00cf2f1892ed73779/episodio/t2e219",
    ]

    should_not_match = [
        "https://tviplayer.iol.pt/programas/Exclusivos",
    ]



from typing import Any, Callable, FrozenSet, List, Optional, Pattern, Sequence, Set, Tuple, Type, Union


class SchemaContainer:
    """
    A simple schema container.
    """

    def __init__(self, schema):
        self.schema = schema


class _CollectionSchemaContainer(SchemaContainer):
    def __init__(self, *schemas):
        super().__init__(schemas)


class AllSchema(_CollectionSchemaContainer):
    """
    Collection of schemas where every schema must be valid.
    The last validation result gets returned.
    """


class AnySchema(_CollectionSchemaContainer):
    """
    Collection of schemas where at least one schema must be valid.
    The first successful validation result gets returned.
    """


class NoneOrAllSchema(_CollectionSchemaContainer):
    """
    Collection of schemas where every schema must be valid. If the initial input is None, all validations will be skipped.
    The last validation result gets returned.
    """


class ListSchema(_CollectionSchemaContainer):
    """
    Collection of schemas where every indexed schema must be valid, as well as the input type and length.
    A new list of the validated input gets returned.
    """


class GetItemSchema:
    """
    Get an item from the input.

    Unless strict is set to True, item can be a tuple of items for recursive lookups.
    If the item is not found in the last object of a recursive lookup, return the default.
    Supported inputs are XML elements, regex matches and anything that implements __getitem__.
    """

    def __init__(
        self,
        item: Union[Any, Tuple[Any]],
        default: Any = None,
        strict: bool = False,
    ):
        self.item = item
        self.default = default
        self.strict = strict


class RegexSchema:
    """
    A regex pattern that must match using the provided method.
    """

    def __init__(
        self,
        pattern: Pattern,
        # TODO: change type from str to Literal["search", "match", "fullmatch", "findall", "split", "sub", "subn"]
        method: str = "search",
    ):
        self.pattern = pattern
        self.method = method


class TransformSchema:
    """
    Transform the input using the specified function and args/keywords.
    """

    def __init__(
        self,
        func: Callable,
        *args,
        **kwargs,
    ):
        self.func = func
        self.args = args
        self.kwargs = kwargs


class OptionalSchema:
    """
    An optional key set in a dict or dict in a :class:`UnionSchema`.
    """

    def __init__(self, key: Any):
        self.key = key


class AttrSchema(SchemaContainer):
    """
    Validate attributes of an input object.
    """


class XmlElementSchema:
    """
    Validate an XML element.
    """

    # signature is weird because of backwards compatiblity
    def __init__(
        self,
        tag: Optional[Any] = None,
        text: Optional[Any] = None,
        attrib: Optional[Any] = None,
        tail: Optional[Any] = None,
    ):
        self.tag = tag
        self.attrib = attrib
        self.text = text
        self.tail = tail


class UnionGetSchema:
    """
    Validate multiple :class:`GetItemSchema` schemas on the same input.
    """

    def __init__(
        self,
        *getters,
        seq: Type[Union[List, FrozenSet, Set, Tuple]] = tuple,
    ):
        self.getters: Sequence[GetItemSchema] = tuple(GetItemSchema(getter) for getter in getters)
        self.seq = seq


class UnionSchema(SchemaContainer):
    """
    Validate multiple schemas on the same input.

    Can be a tuple, list, set, frozenset or dict of schemas.
    """



from datetime import datetime
from os.path import sep
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
from freezegun import freeze_time

from streamlink_cli.utils.formatter import Formatter


class TestCLIFormatter:
    class Obj:
        def __str__(self):
            return "obj"

    @pytest.fixture
    def mock_replace_chars(self):
        with patch("streamlink_cli.utils.formatter.replace_chars") as mock_replace_chars:
            yield mock_replace_chars

    @pytest.fixture
    def prop(self):
        return Mock(return_value="prop")

    @pytest.fixture
    def obj(self):
        return self.Obj()

    @pytest.fixture
    def formatter(self, prop: Mock, obj: Obj):
        with freeze_time("2000-01-02T03:04:05.000006Z"):
            yield Formatter(
                {
                    "prop": prop,
                    "obj": lambda: obj,
                    "time": datetime.now,
                    "empty": lambda: "",
                    "none": lambda: None,
                },
                {
                    "time": lambda dt, fmt: dt.strftime(fmt),
                },
            )

    def test_title(self, formatter: Formatter):
        assert formatter.title == formatter.format

    def test_path(self, formatter: Formatter, prop: Mock, obj: Obj, mock_replace_chars: Mock):
        mock_replace_chars.side_effect = lambda s, *_: s.upper()

        assert formatter.path("text '{prop}' '{empty}' '{none}'") == Path("text 'PROP' '' ''")
        assert formatter.cache == dict(prop="prop", empty="", none=None)
        assert prop.call_count == 1
        assert mock_replace_chars.call_args_list == [
            call("prop", None),
            call("", None),
            call("", None),
        ]
        mock_replace_chars.reset_mock()

        assert formatter.path("text '{prop}' '{obj}' '{empty}' '{none}'", "foo") == Path("text 'PROP' 'OBJ' '' ''")
        assert formatter.cache == dict(prop="prop", obj=obj, empty="", none=None)
        assert prop.call_count == 1
        assert mock_replace_chars.call_args_list == [
            call("prop", "foo"),
            call("obj", "foo"),
            call("", "foo"),
            call("", "foo"),
        ]

    def test_path_substitute(self, formatter: Formatter):
        formatter.mapping.update(**{
            "current": lambda: ".",
            "parent": lambda: "..",
            "dots": lambda: "...",
            "separator": lambda: sep,
        })
        path = formatter.path(f"{{current}}{sep}{{parent}}{sep}{{dots}}{sep}{{separator}}{sep}foo{sep}.{sep}..{sep}bar")
        assert path == Path("_", "_", "...", "_", "foo", ".", "..", "bar"), \
            "Formats the path's parts separately and ignores current and parent directories in substitutions only"



import pytest
import requests_mock as rm

from streamlink import Streamlink
from streamlink.exceptions import NoStreamsError
from streamlink.plugins.vk import VK
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVK(PluginCanHandleUrl):
    __plugin__ = VK

    should_match = [
        "https://vk.com/video-9944999_456239622",
        "http://vk.com/video-24136539_456239830",
        "https://www.vk.com/video-34453259_456240574",
        "https://vk.com/videos-24136539?z=video-24136539_456241155%2Fpl_-24136539_-2",
        "https://vk.com/video?z=video-15755094_456245149%2Fpl_cat_lives",
        "https://vk.com/video?z=video132886594_167211693%2Fpl_cat_8",
        "https://vk.com/video132886594_167211693",
        "https://vk.com/videos132886594?z=video132886594_167211693",
        "https://vk.com/video-73154028_456239128",
        "https://vk.com/dota2?w=wall-126093223_1057924",
        "https://vk.com/search?c%5Bq%5D=dota&c%5Bsection%5D=auto&z=video295508334_171402661",
    ]

    should_not_match = [
        "https://vk.com/",
    ]


@pytest.mark.parametrize("url,newurl", [
    # canonical video URL
    pytest.param(
        "https://vk.com/video-24136539_456241176",
        "https://vk.com/video-24136539_456241176",
    ),

    # video ID from URL
    pytest.param(
        "https://vk.com/videos-24136539?z=video-24136539_456241181%2Fpl_-24136539_-2",
        "https://vk.com/video-24136539_456241181",
    ),
    pytest.param(
        "https://vk.com/videos132886594?z=video132886594_167211693",
        "https://vk.com/video132886594_167211693",
    ),
    pytest.param(
        "https://vk.com/search?c%5Bq%5D=dota&c%5Bsection%5D=auto&z=video295508334_171402661",
        "https://vk.com/video295508334_171402661",
    ),

    # video ID from HTTP request
    pytest.param(
        "https://vk.com/dota2?w=wall-126093223_1057924",
        "https://vk.com/video-126093223_1057924",
    ),

    # no video ID
    pytest.param(
        "https://vk.com/videos-0123456789?z=",
        "",
        marks=pytest.mark.xfail(raises=NoStreamsError),
    ),
    pytest.param(
        "https://vk.com/video/for_kids?z=video%2Fpl_-22277933_51053000",
        "",
        marks=pytest.mark.xfail(raises=NoStreamsError),
    ),
])
def test_url_redirect(url: str, newurl: str, requests_mock):
    session = Streamlink()
    plugin: VK = VK(session, url)
    requests_mock.register_uri(rm.ANY, rm.ANY, exc=rm.exceptions.InvalidRequest)
    requests_mock.get(url, text=f"""<!DOCTYPE html><html><head><meta property="og:url" content="{newurl}"/></head></html>""")
    plugin.follow_vk_redirect()
    assert plugin.url == newurl



from streamlink.plugins.vimeo import Vimeo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVimeo(PluginCanHandleUrl):
    __plugin__ = Vimeo

    should_match = [
        "https://vimeo.com/237163735",
        "https://vimeo.com/channels/music/176894130",
        "https://vimeo.com/album/3706071/video/148903960",
        "https://vimeo.com/ondemand/surveyopenspace/92630739",
        "https://vimeo.com/ondemand/100footsurfingdays",
        "https://player.vimeo.com/video/176894130",
    ]

    should_not_match = [
        "https://www.vimeo.com/"
    ]



"""
$description Global video hosting platform.
$url streamable.com
$type vod
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?streamable\.com/(.+)"
))
class Streamable(Plugin):
    def _get_streams(self):
        data = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"var\s*videoObject\s*=\s*({.*?});"),
            validate.none_or_all(
                validate.get(1),
                validate.parse_json(),
                {
                    "files": {
                        str: {
                            "url": validate.url(),
                            "width": int,
                            "height": int,
                            "bitrate": int,
                        },
                    },
                },
            ),
        ))

        for info in data["files"].values():
            stream_url = update_scheme("https://", info["url"])
            # pick the smaller of the two dimensions, for landscape v. portrait videos
            res = min(info["width"], info["height"])
            yield f"{res}p", HTTPStream(self.session, stream_url)


__plugin__ = Streamable



import unittest

from streamlink.stream.wrappers import StreamIOIterWrapper


class TestPluginStream(unittest.TestCase):
    def test_iter(self):
        def generator():
            yield b"1" * 8192
            yield b"2" * 4096
            yield b"3" * 2048

        fd = StreamIOIterWrapper(generator())
        self.assertEqual(fd.read(4096), b"1" * 4096)
        self.assertEqual(fd.read(2048), b"1" * 2048)
        self.assertEqual(fd.read(2048), b"1" * 2048)
        self.assertEqual(fd.read(1), b"2")
        self.assertEqual(fd.read(4095), b"2" * 4095)
        self.assertEqual(fd.read(1536), b"3" * 1536)
        self.assertEqual(fd.read(), b"3" * 512)



"""
$description Live TV channels and video on-demand service from TV4, a Swedish free-to-air broadcaster.
$url tv4play.se
$url fotbollskanalen.se
$type live, vod
$region Sweden
$notes Only non-premium streams are supported
"""

import logging
import re
from urllib.parse import urljoin

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?
    (?:
        tv4play\.se/program/[^?/]+/[^?/]+
        |
        fotbollskanalen\.se/video
    )
    /(?P<video_id>\d+)
""", re.VERBOSE))
class TV4Play(Plugin):
    video_id = None

    api_url = "https://playback-api.b17g.net"
    api_assets = urljoin(api_url, "/asset/{0}")

    _meta_schema = validate.Schema(
        {
            "metadata": {
                "title": str,
            },
            "mediaUri": str,
        }
    )

    @property
    def get_video_id(self):
        if self.video_id is None:
            self.video_id = self.match.group("video_id")
            log.debug("Found video ID: {0}".format(self.video_id))
        return self.video_id

    def get_metadata(self):
        params = {
            "device": "browser",
            "protocol": "hls",
            "service": "tv4",
        }
        try:
            res = self.session.http.get(
                self.api_assets.format(self.get_video_id),
                params=params
            )
        except Exception as e:
            if "404 Client Error" in str(e):
                raise PluginError("This Video is not available")
            raise e
        log.debug("Found metadata")
        metadata = self.session.http.json(res, schema=self._meta_schema)
        self.title = metadata["metadata"]["title"]
        return metadata

    def get_title(self):
        if self.title is None:
            self.get_metadata()
        return self.title

    def _get_streams(self):
        metadata = self.get_metadata()

        try:
            res = self.session.http.get(urljoin(self.api_url, metadata["mediaUri"]))
        except Exception as e:
            if "401 Client Error" in str(e):
                raise PluginError("This Video is not available in your country")
            raise e

        log.debug("Found stream data")
        data = self.session.http.json(res)
        hls_url = data["playbackItem"]["manifestUrl"]
        log.debug("URL={0}".format(hls_url))
        yield from HLSStream.parse_variant_playlist(self.session, hls_url).items()


__plugin__ = TV4Play



"""
$description Live TV channels and video on-demand service from SVT, a Swedish public, state-owned broadcaster.
$url svtplay.se
$type live, vod
$region Sweden
"""

import logging
import re
from urllib.parse import parse_qsl, urlparse

from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.ffmpegmux import MuxedStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?svtplay\.se/(?P<live>kanaler/)?"
))
@pluginargument(
    "mux-subtitles",
    is_global=True,
)
class SVTPlay(Plugin):
    _URL_API_VIDEO = "https://api.svt.se/videoplayer-api/video/{item}"
    _MAP_CHANNEL_NAMES = {
        "svt1": "ch-svt1",
        "svt2": "ch-svt2",
        "svtbarn": "ch-barnkanalen",
        "kunskapskanalen": "ch-kunskapskanalen",
        "svt24": "ch-svt24",
    }

    def _api_call(self, item):
        _schema_items = validate.all(
            [
                validate.all(
                    {
                        "format": str,
                        "url": validate.url(),
                    },
                    validate.union_get("format", "url"),
                ),
            ],
            validate.transform(dict),
        )

        self.author, self.title, videos, subtitles = self.session.http.get(
            self._URL_API_VIDEO.format(item=item),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("programTitle"): str,
                    validate.optional("episodeTitle"): str,
                    "videoReferences": _schema_items,
                    validate.optional("subtitleReferences"): _schema_items,
                },
                validate.union_get(
                    "programTitle",
                    "episodeTitle",
                    "videoReferences",
                    "subtitleReferences",
                ),
            ),
        )

        return videos, subtitles

    def _get_live(self):
        live_id = "/".join(urlparse(self.url).path.split("/")[2:])
        if not live_id:
            return

        live_id = self._MAP_CHANNEL_NAMES.get(live_id, f"ch-{live_id}")
        log.debug(f"Live ID={live_id}")
        self.category = "Live"
        videos, subtitles = self._api_call(live_id)

        return self._select_streams(videos, subtitles)

    def _get_vod(self):
        def get_vod_id(url):
            return dict(parse_qsl(urlparse(url).query)).get("id")

        vod_id = get_vod_id(self.url)

        if vod_id is None:
            vod_url = self.session.http.get(self.url, schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//*[@data-rt='top-area-play-button'][@href][1]/@href"),
            ))
            if vod_url:
                vod_id = get_vod_id(vod_url)

        if vod_id is None:
            return

        log.debug(f"VOD ID={vod_id}")
        self.category = "Live"
        videos, subtitles = self._api_call(vod_id)

        return self._select_streams(videos, subtitles)

    def _select_streams(self, videos, subtitles):
        # the goal is to have streams with the widest range of qualities/substreams and highest bitrate at the top
        stream_priorities = {
            "dashhbbtv": DASHStream,  # DASH AVC
            "dash-hbbtv-avc": DASHStream,  # DASH AVC
            "dash-avc": DASHStream,  # DASH AVC
            "dash-full": DASHStream,  # DASH AVC
            "dash": DASHStream,  # DASH AVC

            "hlswebvtt": HLSStream,  # HLS with subtitles
            "hls-cmaf-live-vtt": HLSStream,  # HLS with subtitles
            "hls-ts-avc": HLSStream,  # HLS with MPEG-TS
            "hls-ts-full": HLSStream,  # HLS with MPEG-TS
            "hls": HLSStream,  # HLS with MPEG-TS
            "hls-cmaf-live": HLSStream,  # HLS with fMP4
            "hls-cmaf-full": HLSStream,  # HLS with fMP4

            "dash-hbbtv-hevc": DASHStream,  # DASH HEVC (low prio, because of potential user decoder issues)

            "hls-ts-lb-full": HLSStream,  # low bitrate
            "hls-cmaf-lb-full": HLSStream,  # low bitrate
            "dash-lb-full": DASHStream,  # low bitrate
        }

        for fmt, streamtype in stream_priorities.items():
            if fmt not in videos:
                continue

            if streamtype is HLSStream:
                return HLSStream.parse_variant_playlist(self.session, videos[fmt], name_fmt="{pixels}_{bitrate}")

            if streamtype is DASHStream:
                mux_subtitles = self.get_option("mux_subtitles")
                subtitlestreams = {}
                if mux_subtitles and "webvtt" in subtitles:
                    subtitlestreams["webvtt"] = HTTPStream(self.session, subtitles["webvtt"])

                dash_streams = DASHStream.parse_manifest(self.session, videos[fmt])
                if not subtitlestreams:
                    return dash_streams

                return {q: MuxedStream(self.session, s, subtitles=subtitlestreams) for q, s in dash_streams.items()}

    def _get_streams(self):
        if self.match["live"]:
            return self._get_live()
        else:
            return self._get_vod()


__plugin__ = SVTPlay



from collections import abc
from copy import copy, deepcopy
from functools import singledispatch
from re import Match, Pattern

from lxml.etree import Element, iselement

from streamlink.exceptions import PluginError
from streamlink.plugin.api.validate._exception import ValidationError
from streamlink.plugin.api.validate._schemas import (
    AllSchema,
    AnySchema,
    AttrSchema,
    GetItemSchema,
    ListSchema,
    NoneOrAllSchema,
    OptionalSchema,
    RegexSchema,
    TransformSchema,
    UnionGetSchema,
    UnionSchema,
    XmlElementSchema,
)


class Schema(AllSchema):
    """
    Wrapper class for :class:`AllSchema` with a validate method which raises :class:`PluginError` by default on error.
    """

    def validate(self, value, name="result", exception=PluginError):
        try:
            return validate(self, value)
        except ValidationError as err:
            raise exception(f"Unable to validate {name}: {err}") from None


# ----


@singledispatch
def validate(schema, value):
    if schema != value:
        raise ValidationError(
            "{value} does not equal {expected}",
            value=repr(value),
            expected=repr(schema),
            schema="equality",
        )

    return value


@validate.register(type)
def _validate_type(schema, value):
    if not isinstance(value, schema):
        raise ValidationError(
            "Type of {value} should be {expected}, but is {actual}",
            value=repr(value),
            expected=schema.__name__,
            actual=type(value).__name__,
            schema=type,
        )

    return value


@validate.register(list)
@validate.register(tuple)
@validate.register(set)
@validate.register(frozenset)
def _validate_sequence(schema, value):
    cls = type(schema)
    validate(cls, value)

    return cls(
        validate(AnySchema(*schema), v) for v in value
    )


@validate.register(dict)
def _validate_dict(schema, value):
    cls = type(schema)
    validate(cls, value)
    new = cls()

    for key, subschema in schema.items():
        if isinstance(key, OptionalSchema):
            if key.key not in value:
                continue
            key = key.key

        if type(key) in (type, AllSchema, AnySchema, TransformSchema, UnionSchema):
            for subkey, subvalue in value.items():
                try:
                    newkey = validate(key, subkey)
                except ValidationError as err:
                    raise ValidationError("Unable to validate key", schema=dict) from err
                try:
                    newvalue = validate(subschema, subvalue)
                except ValidationError as err:
                    raise ValidationError("Unable to validate value", schema=dict) from err
                new[newkey] = newvalue
            break

        if key not in value:
            raise ValidationError(
                "Key {key} not found in {value}",
                key=repr(key),
                value=repr(value),
                schema=dict,
            )

        try:
            new[key] = validate(subschema, value[key])
        except ValidationError as err:
            raise ValidationError("Unable to validate value of key {key}", key=repr(key), schema=dict) from err

    return new


@validate.register
def _validate_callable(schema: abc.Callable, value):
    if not schema(value):
        raise ValidationError(
            "{callable} is not true",
            callable=f"{schema.__name__}({value!r})",
            schema=abc.Callable,
        )

    return value


@validate.register
def _validate_pattern(schema: Pattern, value):
    if not isinstance(value, (str, bytes)):
        raise ValidationError(
            "Type of {value} should be str or bytes, but is {actual}",
            value=repr(value),
            actual=type(value).__name__,
            schema=Pattern,
        )

    try:
        result = schema.search(value)
    except TypeError as err:
        raise ValidationError(err, schema=Pattern) from None

    return result


@validate.register
def _validate_allschema(schema: AllSchema, value):
    for schema in schema.schema:
        value = validate(schema, value)

    return value


@validate.register
def _validate_anyschema(schema: AnySchema, value):
    errors = []
    for subschema in schema.schema:
        try:
            return validate(subschema, value)
        except ValidationError as err:
            errors.append(err)

    raise ValidationError(*errors, schema=AnySchema)


@validate.register
def _validate_noneorallschema(schema: NoneOrAllSchema, value):
    if value is not None:
        try:
            for schema in schema.schema:
                value = validate(schema, value)
        except ValidationError as err:
            raise ValidationError(err, schema=NoneOrAllSchema) from None

    return value


@validate.register
def _validate_listschema(schema: ListSchema, value):
    if type(value) is not list:
        raise ValidationError(
            "Type of {value} should be list, but is {actual}",
            value=repr(value),
            actual=type(value).__name__,
            schema=ListSchema,
        )
    if len(value) != len(schema.schema):
        raise ValidationError(
            "Length of list ({length}) does not match expectation ({expected})",
            length=len(value),
            expected=len(schema.schema),
            schema=ListSchema,
        )

    new = []
    errors = []
    for k, v in enumerate(schema.schema):
        try:
            new.append(validate(v, value[k]))
        except ValidationError as err:
            errors.append(err)

    if errors:
        raise ValidationError(*errors, schema=ListSchema)

    return new


@validate.register
def _validate_regexschema(schema: RegexSchema, value):
    if not isinstance(value, (str, bytes)):
        raise ValidationError(
            "Type of {value} should be str or bytes, but is {actual}",
            value=repr(value),
            actual=type(value).__name__,
            schema=RegexSchema,
        )

    try:
        result = getattr(schema.pattern, schema.method)(value)
    except TypeError as err:
        raise ValidationError(err, schema=RegexSchema) from None

    if result is None:
        raise ValidationError(
            "Pattern {pattern} did not match {value}",
            pattern=repr(schema.pattern.pattern),
            value=repr(value),
            schema=RegexSchema,
        )

    return result


@validate.register
def _validate_transformschema(schema: TransformSchema, value):
    validate(abc.Callable, schema.func)
    return schema.func(value, *schema.args, **schema.kwargs)


@validate.register
def _validate_getitemschema(schema: GetItemSchema, value):
    item = schema.item if type(schema.item) is tuple and not schema.strict else (schema.item,)
    idx = 0
    key = None
    try:
        for key in item:
            if iselement(value):
                value = value.attrib[key]
            elif isinstance(value, Match):
                value = value.group(key)
            else:
                value = value[key]
            idx += 1
        return value
    except (KeyError, IndexError):
        # only return default value on last item in nested lookup
        if idx < len(item) - 1:
            raise ValidationError(
                "Item {key} was not found in object {value}",
                key=repr(key),
                value=repr(value),
                schema=GetItemSchema,
            ) from None
        return schema.default
    except (TypeError, AttributeError) as err:
        raise ValidationError(
            "Could not get key {key} from object {value}",
            key=repr(key),
            value=repr(value),
            schema=GetItemSchema,
        ) from err


@validate.register
def _validate_attrschema(schema: AttrSchema, value):
    new = copy(value)

    for key, subschema in schema.schema.items():
        if not hasattr(value, key):
            raise ValidationError(
                "Attribute {key} not found on object {value}",
                key=repr(key),
                value=repr(value),
                schema=AttrSchema,
            )

        try:
            value = validate(subschema, getattr(value, key))
        except ValidationError as err:
            raise ValidationError(
                "Could not validate attribute {key}",
                key=repr(key),
                schema=AttrSchema,
            ) from err

        setattr(new, key, value)

    return new


@validate.register
def _validate_xmlelementschema(schema: XmlElementSchema, value):
    validate(iselement, value)
    tag = value.tag
    attrib = value.attrib
    text = value.text
    tail = value.tail

    if schema.tag is not None:
        try:
            tag = validate(schema.tag, value.tag)
        except ValidationError as err:
            raise ValidationError("Unable to validate XML tag", schema=XmlElementSchema) from err

    if schema.attrib is not None:
        try:
            attrib = validate(schema.attrib, dict(value.attrib))
        except ValidationError as err:
            raise ValidationError("Unable to validate XML attributes", schema=XmlElementSchema) from err

    if schema.text is not None:
        try:
            text = validate(schema.text, value.text)
        except ValidationError as err:
            raise ValidationError("Unable to validate XML text", schema=XmlElementSchema) from err

    if schema.tail is not None:
        try:
            tail = validate(schema.tail, value.tail)
        except ValidationError as err:
            raise ValidationError("Unable to validate XML tail", schema=XmlElementSchema) from err

    new = Element(tag, attrib)
    new.text = text
    new.tail = tail
    for child in value:
        new.append(deepcopy(child))

    return new


@validate.register
def _validate_uniongetschema(schema: UnionGetSchema, value):
    return schema.seq(
        validate(getter, value) for getter in schema.getters
    )


@validate.register
def _validate_unionschema(schema: UnionSchema, value):
    try:
        return validate_union(schema.schema, value)
    except ValidationError as err:
        raise ValidationError("Could not validate union", schema=UnionSchema) from err


# ----


# noinspection PyUnusedLocal
@singledispatch
def validate_union(schema, value):
    raise ValidationError(
        "Invalid union type: {type}",
        type=type(schema).__name__,
    )


@validate_union.register(dict)
def _validate_union_dict(schema, value):
    new = type(schema)()
    for key, schema in schema.items():
        is_optional = isinstance(key, OptionalSchema)
        if is_optional:
            key = key.key

        try:
            new[key] = validate(schema, value)
        except ValidationError as err:
            if is_optional:
                continue

            raise ValidationError(
                "Unable to validate union {key}",
                key=repr(key),
                schema=dict,
            ) from err

    return new


@validate_union.register(list)
@validate_union.register(tuple)
@validate_union.register(set)
@validate_union.register(frozenset)
def _validate_union_sequence(schemas, value):
    return type(schemas)(
        validate(schema, value) for schema in schemas
    )



from streamlink.plugins.sbscokr import SBScokr
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSBScokr(PluginCanHandleUrl):
    __plugin__ = SBScokr

    should_match_groups = [
        # regular channels
        ("https://www.sbs.co.kr/live/S01", {"channel": "S01"}),
        ("https://www.sbs.co.kr/live/S01?div=live_list", {"channel": "S01"}),
        # "virtual" channels
        ("https://www.sbs.co.kr/live/S21", {"channel": "S21"}),
        ("https://www.sbs.co.kr/live/S21?div=live_list", {"channel": "S21"}),
    ]



"""
$description German sports magazine live stream, owned by ARD.
$url sportschau.de
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:\w+\.)*sportschau\.de/"
))
class Sportschau(Plugin):
    def _get_streams(self):
        player_js = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"https?:(//deviceids-medp.wdr.de/ondemand/\S+\.js)"),
            validate.none_or_all(
                validate.get(1),
                validate.transform(lambda url: update_scheme("https://", url)),
            ),
        ))
        if not player_js:
            return

        log.debug(f"Found player js {player_js}")
        data = self.session.http.get(player_js, schema=validate.Schema(
            validate.regex(re.compile(r"\$mediaObject\.jsonpHelper\.storeAndPlay\(({.+})\);?")),
            validate.get(1),
            validate.parse_json(),
            validate.get("mediaResource"),
            validate.get("dflt"),
            {
                validate.optional("audioURL"): validate.url(),
                validate.optional("videoURL"): validate.url(),
            },
        ))

        if data.get("videoURL"):
            yield from HLSStream.parse_variant_playlist(self.session, update_scheme("https:", data.get("videoURL"))).items()
        if data.get("audioURL"):
            yield "audio", HTTPStream(self.session, update_scheme("https:", data.get("audioURL")))


__plugin__ = Sportschau



from streamlink.plugins.btv import BTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBTV(PluginCanHandleUrl):
    __plugin__ = BTV

    should_match = [
        "http://btvplus.bg/live",
        "http://btvplus.bg/live/",
        "http://www.btvplus.bg/live/",
    ]



"""
$description Live TV channels from VTV, a Vietnamese public, state-owned broadcaster.
$url vtvgo.vn
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://vtvgo\.vn/xem-truc-tuyen-kenh-"
))
class VTVgo(Plugin):
    AJAX_URL = "https://vtvgo.vn/ajax-get-stream"

    def _get_streams(self):
        # get cookies
        self.session.http.get("https://vtvgo.vn/")

        self.session.http.headers.update({
            "Origin": "https://vtvgo.vn",
            "Referer": self.url,
            "Sec-Fetch-Site": "same-origin",
            "X-Requested-With": "XMLHttpRequest",
        })

        params = self.session.http.get(self.url, schema=validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//script[contains(text(),'setplayer(')][1]/text()"),
            validate.none_or_all(
                validate.regex(
                    re.compile(r"""var\s+(?P<key>(?:type_)?id|time|token)\s*=\s*["']?(?P<value>[^"']+)["']?;"""),
                    method="findall",
                ),
                [
                    ("id", int),
                    ("type_id", str),
                    ("time", str),
                    ("token", str),
                ],
            ),
        ))
        if not params:
            return

        log.trace(f"{params!r}")
        hls_url = self.session.http.post(
            self.AJAX_URL,
            data=dict(params),
            schema=validate.Schema(
                validate.parse_json(),
                {"stream_url": [validate.url()]},
                validate.get(("stream_url", 0)),
            ),
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = VTVgo



import unittest
from unittest.mock import Mock, mock_open, patch

from streamlink import Streamlink
from streamlink.stream.file import FileStream


class TestFileStream(unittest.TestCase):
    def setUp(self):
        self.session = Streamlink()

    def test_open_file_path(self):
        m = mock_open()
        s = FileStream(self.session, path="/test/path")
        with patch('streamlink.stream.file.open', m, create=True):
            s.open()
            m.assert_called_with("/test/path")

    def test_open_fileobj(self):
        fileobj = Mock()
        s = FileStream(self.session, fileobj=fileobj)
        self.assertEqual(fileobj, s.open())



"""
$description Turkish live TV channel owned by Galatasaray TV.
$url galatasaray.com
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?galatasaray\.com"
))
class GalatasarayTV(Plugin):
    playervars_re = re.compile(r"sources\s*:\s*\[\s*\{\s*type\s*:\s*\"(.*?)\",\s*src\s*:\s*\"(.*?)\"", re.DOTALL)

    title = "Galatasaray TV"

    def _get_streams(self):
        res = self.session.http.get(self.url)
        match = self.playervars_re.search(res.text)
        if match:
            stream_url = match.group(2)
            log.debug("URL={0}".format(stream_url))
            return HLSStream.parse_variant_playlist(self.session, stream_url)


__plugin__ = GalatasarayTV



from streamlink.plugins.qq import QQ
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlQQ(PluginCanHandleUrl):
    __plugin__ = QQ

    should_match_groups = [
        ("https://live.qq.com/", {}),
        ("https://live.qq.com/123456789", {"room_id": "123456789"}),
        ("https://m.live.qq.com/123456789", {"room_id": "123456789"}),
    ]



import unittest
from argparse import ArgumentTypeError

from streamlink.utils.args import (
    boolean, comma_list, comma_list_filter, filesize, keyvalue, num
)


class TestUtilsArgs(unittest.TestCase):
    def test_boolean_true(self):
        self.assertEqual(boolean('1'), True)
        self.assertEqual(boolean('on'), True)
        self.assertEqual(boolean('true'), True)
        self.assertEqual(boolean('yes'), True)
        self.assertEqual(boolean('Yes'), True)

    def test_boolean_false(self):
        self.assertEqual(boolean('0'), False)
        self.assertEqual(boolean('false'), False)
        self.assertEqual(boolean('no'), False)
        self.assertEqual(boolean('No'), False)
        self.assertEqual(boolean('off'), False)

    def test_boolean_error(self):
        with self.assertRaises(ArgumentTypeError):
            boolean('yesno')

        with self.assertRaises(ArgumentTypeError):
            boolean('FOO')

        with self.assertRaises(ArgumentTypeError):
            boolean('2')

    def test_comma_list(self):
        # (values, result)
        test_data = [
            ('foo.bar,example.com', ['foo.bar', 'example.com']),
            ('/var/run/foo,/var/run/bar', ['/var/run/foo', '/var/run/bar']),
            ('foo bar,24', ['foo bar', '24']),
            ('hls', ['hls']),
        ]

        for _v, _r in test_data:
            self.assertEqual(comma_list(_v), _r)

    def test_comma_list_filter(self):
        # (acceptable, values, result)
        test_data = [
            (['foo', 'bar', 'com'], 'foo,bar,example.com', ['foo', 'bar']),
            (['/var/run/foo', 'FO'], '/var/run/foo,/var/run/bar',
             ['/var/run/foo']),
            (['hls', 'hls5', 'dash'], 'hls,hls5', ['hls', 'hls5']),
            (['EU', 'RU'], 'DE,FR,RU,US', ['RU']),
        ]

        for _a, _v, _r in test_data:
            func = comma_list_filter(_a)
            self.assertEqual(func(_v), _r)

    def test_filesize(self):
        self.assertEqual(filesize('2000'), 2000)
        self.assertEqual(filesize('11KB'), 1024 * 11)
        self.assertEqual(filesize('12MB'), 1024 * 1024 * 12)
        self.assertEqual(filesize('1KB'), 1024)
        self.assertEqual(filesize('1MB'), 1024 * 1024)
        self.assertEqual(filesize('2KB'), 1024 * 2)

    def test_filesize_error(self):
        with self.assertRaises(ValueError):
            filesize('FOO')

        with self.assertRaises(ValueError):
            filesize('0.00000')

    def test_keyvalue(self):
        # (value, result)
        test_data = [
            ('X-Forwarded-For=127.0.0.1', ('X-Forwarded-For', '127.0.0.1')),
            ('Referer=https://foo.bar', ('Referer', 'https://foo.bar')),
            (
                'User-Agent=Mozilla/5.0 (X11; Linux x86_64; rv:60.0)'
                ' Gecko/20100101 Firefox/60.0',
                ('User-Agent', 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) '
                 'Gecko/20100101 Firefox/60.0')
            ),
            ('domain=example.com', ('domain', 'example.com')),
        ]

        for _v, _r in test_data:
            self.assertEqual(keyvalue(_v), _r)

    def test_keyvalue_error(self):
        with self.assertRaises(ValueError):
            keyvalue('127.0.0.1')

    def test_num(self):
        # (value, func, result)
        test_data = [
            ('33', num(int, 5, 120), 33),
            ('234', num(int, min=10), 234),
            ('50.222', num(float, 10, 120), 50.222),
        ]

        for _v, _f, _r in test_data:
            self.assertEqual(_f(_v), _r)

    def test_num_error(self):
        with self.assertRaises(ArgumentTypeError):
            func = num(int, 5, 10)
            func('3')

        with self.assertRaises(ArgumentTypeError):
            func = num(int, max=11)
            func('12')

        with self.assertRaises(ArgumentTypeError):
            func = num(int, min=15)
            func('8')

        with self.assertRaises(ArgumentTypeError):
            func = num(float, 10, 20)
            func('40.222')



"""
$description Global live streaming platform.
$url ssh101.com
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?ssh101\.com/(?:(?:secure)?live/|detail\.php\?id=\w+)"
))
class SSH101(Plugin):
    def _get_streams(self):
        hls_url = self.session.http.get(self.url, schema=validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//source[contains(@src,'.m3u8')]/@src"),
        ))
        if not hls_url:
            return

        res = self.session.http.get(hls_url, acceptable_status=(200, 403, 404))
        if res.status_code != 200 or len(res.text) <= 10:
            log.error("This stream is currently offline")
            return

        return {"live": HLSStream(self.session, hls_url)}


__plugin__ = SSH101



"""
$description European public service channel promoting culture, including magazine shows, concerts and documentaries.
$url arte.tv
$type live, vod
"""

import logging
import re
from operator import itemgetter

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:\w+\.)?arte\.tv/(?:guide/)?
    (?P<language>[a-z]{2})/
    (?:
        (?:videos/)?(?P<video_id>(?!RC-|videos)[^/]+?)/.+
        |
        (?:direct|live)
    )
""", re.VERBOSE))
class ArteTV(Plugin):
    API_URL = "https://api.arte.tv/api/player/v2/config/{0}/{1}"
    API_TOKEN = "MzYyZDYyYmM1Y2Q3ZWRlZWFjMmIyZjZjNTRiMGY4MzY4NzBhOWQ5YjE4MGQ1NGFiODJmOTFlZDQwN2FkOTZjMQ"

    def _get_streams(self):
        language = self.match.group("language")
        video_id = self.match.group("video_id")

        json_url = self.API_URL.format(language, video_id or "LIVE")
        headers = {
            "Authorization": f"Bearer {self.API_TOKEN}"
        }
        streams, metadata = self.session.http.get(json_url, headers=headers, schema=validate.Schema(
            validate.parse_json(),
            {"data": {"attributes": {
                "streams": validate.any(
                    [],
                    [
                        validate.all(
                            {
                                "url": validate.url(),
                                "slot": int,
                                "protocol": validate.any("HLS", "HLS_NG"),
                            },
                            validate.union_get("slot", "protocol", "url")
                        )
                    ]
                ),
                "metadata": {
                    "title": str,
                    "subtitle": validate.any(None, str)
                }
            }}},
            validate.get(("data", "attributes")),
            validate.union_get("streams", "metadata")
        ))

        if not streams:
            return

        self.title = f"{metadata['title']} - {metadata['subtitle']}" if metadata["subtitle"] else metadata["title"]

        for slot, protocol, url in sorted(streams, key=itemgetter(0)):
            return HLSStream.parse_variant_playlist(self.session, url)


__plugin__ = ArteTV



import datetime
from unittest.mock import patch

import freezegun
import pytest

from streamlink import Streamlink
from streamlink.plugins.filmon import FilmOnAPI, FilmOnHLS, Filmon
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlFilmon(PluginCanHandleUrl):
    __plugin__ = Filmon

    should_match = [
        'http://www.filmon.tv/channel/grandstand-show',
        'http://www.filmon.tv/index/popout?channel_id=5510&quality=low',
        'http://www.filmon.tv/tv/channel/export?channel_id=5510&autoPlay=1',
        'http://www.filmon.tv/tv/channel/grandstand-show',
        'http://www.filmon.tv/tv/channel-4',
        'https://www.filmon.com/tv/bbc-news',
        'https://www.filmon.tv/tv/55',
        'http://www.filmon.tv/vod/view/10250-0-crime-boss',
        'http://www.filmon.tv/group/comedy',
    ]

    should_match_groups = [
        ('http://www.filmon.tv/channel/grandstand-show', (None, "grandstand-show", None)),
        ('http://www.filmon.tv/index/popout?channel_id=5510&quality=low', (None, '5510', None)),
        ('http://www.filmon.tv/tv/channel/export?channel_id=5510&autoPlay=1', (None, '5510', None)),
        ('http://www.filmon.tv/tv/channel/grandstand-show', (None, 'grandstand-show', None)),
        ('https://www.filmon.com/tv/bbc-news', (None, 'bbc-news', None)),
        ('https://www.filmon.com/tv/channel-4', (None, 'channel-4', None)),
        ('https://www.filmon.tv/tv/55', (None, '55', None)),
        ('http://www.filmon.tv/group/comedy', ('group/', 'comedy', None)),
        ('http://www.filmon.tv/vod/view/10250-0-crime-boss', (None, None, '10250-0-crime-boss')),
        ('http://www.filmon.tv/vod/view/10250-0-crime-boss/extra', (None, None, '10250-0-crime-boss')),
        ('http://www.filmon.tv/vod/view/10250-0-crime-boss?extra', (None, None, '10250-0-crime-boss')),
        ('http://www.filmon.tv/vod/view/10250-0-crime-boss&extra', (None, None, '10250-0-crime-boss')),
    ]


@pytest.fixture(scope="function")
def filmonhls():
    with freezegun.freeze_time(datetime.datetime(2000, 1, 1, 0, 0, 0, 0)), \
         patch("streamlink.plugins.filmon.FilmOnHLS._get_stream_data", return_value=[]):
        session = Streamlink()
        api = FilmOnAPI(session)
        yield FilmOnHLS(session, "http://fake/one.m3u8", api=api, channel="test")


def test_filmonhls_to_url(filmonhls):
    filmonhls.watch_timeout = datetime.datetime(2000, 1, 1, 0, 0, 0, 0).timestamp()
    assert filmonhls.to_url() == "http://fake/one.m3u8"


def test_filmonhls_to_url_updated(filmonhls):
    filmonhls.watch_timeout = datetime.datetime(1999, 12, 31, 23, 59, 59, 9999).timestamp()

    filmonhls._get_stream_data.return_value = [
        ("high", "http://fake/two.m3u8", datetime.datetime(2000, 1, 1, 0, 0, 0, 0).timestamp()),
    ]
    assert filmonhls.to_url() == "http://fake/two.m3u8"

    filmonhls.watch_timeout = datetime.datetime(1999, 12, 31, 23, 59, 59, 9999).timestamp()
    filmonhls._get_stream_data.return_value = [
        ("high", "http://another-fake/three.m3u8", datetime.datetime(2000, 1, 1, 0, 0, 0, 0).timestamp()),
    ]
    assert filmonhls.to_url() == "http://fake/three.m3u8"


def test_filmonhls_to_url_missing_quality(filmonhls):
    filmonhls.watch_timeout = datetime.datetime(1999, 12, 31, 23, 59, 59, 9999).timestamp()

    filmonhls._get_stream_data.return_value = [
        ("low", "http://fake/two.m3u8", datetime.datetime(2000, 1, 1, 0, 0, 0, 0).timestamp()),
    ]
    with pytest.raises(TypeError) as cm:
        filmonhls.to_url()
    assert str(cm.value) == "Stream has expired and cannot be translated to a URL"



from streamlink.plugins.yupptv import YuppTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlYuppTV(PluginCanHandleUrl):
    __plugin__ = YuppTV

    should_match = [
        'https://www.yupptv.com/channels/etv-telugu/live',
        'https://www.yupptv.com/channels/india-today-news/news/25326023/15-jun-2018',
    ]
